#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time

from sqlalchemy import text
from sqlalchemy.exc import MultipleResultsFound, NoResultFound, DBAPIError
from sqlalchemy import and_

from poppy.core.logger import logger

from roc.idb.models.idb import IdbRelease
from roc.idb.constants import TRYOUTS, TIME_WAIT_SEC

__all__ = ["actual_sql", "get_current_idb", "clean_idb", "delete_idb_schema"]


def actual_sql(sqlalchemy_query):
    """
    convert input Sqlalchemy query into explicit SQL syntax query

    :param sqlalchemy_query: input Sqlalchemy query object
    :return: string with corresponding SQL syntax
    """
    return str(
        sqlalchemy_query.statement.compile(compile_kwargs={"literal_binds": True})
    )


def get_current_idb(idb_source, session, tryouts=TRYOUTS, wait=TIME_WAIT_SEC):
    """
    Get current idb release stored in the database

    :param idb_source: IDB source to use (MIB, SRDB or PALISADE).
    :param session: database session
    :param tryouts: number of tries
    :param wait: seconds to wait between two tries
    :return: version of the idb tagged as current, None if not found
    """
    idb_version = None

    filters = []
    if idb_source:
        filters.append(IdbRelease.idb_source == idb_source)
    filters.append(IdbRelease.current is True)
    for i in range(tryouts):
        try:
            query = session.query(IdbRelease.idb_version).filter(and_(*filters))
            results = query.one()
        except MultipleResultsFound:
            logger.error(f"Multiple results found for {actual_sql(query)}!")
            break
        except NoResultFound:
            logger.info(f"No result found for {actual_sql(query)}")
        except Exception as e:
            logger.exception(
                f"Cannot run {actual_sql(query)} (trying again in {wait} seconds)"
            )
            logger.debug(e)
            time.sleep(wait)
        else:
            idb_version = results.idb_version
            break

    return idb_version


def clean_idb(session, idb_source, idb_version):
    """
    Clean a given IDB release in the database

    :param session: database session instance
    :param idb_source: IDB source
    :param idb_version: IDB version
    :return: True if IDB release has been deleted correctly
    """
    is_deleted = False
    try:
        idb_to_delete = (
            session.query(IdbRelease)
            .filter(
                IdbRelease.idb_source == idb_source,
                IdbRelease.idb_version == idb_version,
            )
            .one()
        )
        logger.debug(f"Deleting pre-existing IDB [{idb_source}-{idb_version}] ...")
        session.delete(idb_to_delete)
    except MultipleResultsFound:
        logger.exception(f"Multiple IDB [{idb_source}-{idb_version}] has been found!")
    except NoResultFound:
        is_deleted = True
    except Exception as e:
        logger.exception(
            f"Pre-existing IDB [{idb_source}-{idb_version}] cannot be deleted!"
        )
        logger.debug(e)
    else:
        is_deleted = True

    return is_deleted


def delete_idb_schema(session):
    """
    Drop IDB schema in the database.

    :param session: database engine session
    :return: True if IDB schema has been deleted
    """
    is_deleted = False
    try:
        session.execute(text("DROP SCHEMA IF EXISTS idb CASCADE"))
    except DBAPIError:
        logger.exception("Cannot drop schema idb in the database")
    else:
        logger.debug("Schema idb is dropped")
        is_deleted = True

    return is_deleted



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .enumeration_parser import EnumerationParser
from .packet_parser import PacketParser
from .parameter_parser import ParameterParser
from .tf_parser import TFParser
from .base_palisade_parser import BasePALISADEParser

__all__ = [
    "PALISADEParser",
]


class PALISADEParser(
    BasePALISADEParser,
    ParameterParser,
    EnumerationParser,
    PacketParser,
    TFParser,
):
    """
    Class used for the parsing of the XML of the IDB (PALISADE) to get information on the
    IDB and SRDB.
    """

    def parse(self):
        """
        Call all methods to the necessary information from the IDB and SRDB
        mapping in XML files specified in arguments.
        """

        # create the mapping for packets
        self.get_structures()

        # populate enumerations
        self.populate_enumerations()

        # populate information of parameters
        self.populate_parameters()

        # create packets
        self.create_packets_idb()

        # create TF packets
        self.create_tf()



# -*- coding: utf-8 -*-
"""roc_idb_0001_initial

Revision ID: roc_idb_0001_initial
Revises:
Create Date: 2019-12-26 18:29:52.527835

"""

import sqlalchemy as sa
from poppy.pop.alembic.helpers import (
    create_table,
    create_schema,
    execute,
    user,
    drop_table,
)
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "roc_idb_0001_initial"
down_revision = None
branch_labels = ("idb",)
depends_on = None


def upgrade():
    create_schema("idb")  # added by hand

    ## delete the types if they exist
    execute("DROP TYPE IF EXISTS idb.idb_source_type")

    # ### commands auto generated by Alembic - please adjust! ###
    create_table(
        "idb_release",
        sa.Column("id_idb_release", sa.INTEGER(), nullable=False),
        sa.Column("idb_version", sa.String(length=128), nullable=False),
        sa.Column(
            "idb_source",
            postgresql.ENUM(
                "PALISADE", "MIB", "SRDB", name="idb_source_type", schema="idb"
            ),
            nullable=False,
        ),
        sa.Column("mib_version", sa.String(), nullable=True),
        sa.Column("current", sa.BOOLEAN(), nullable=False),
        sa.Column("validity_min", postgresql.TIMESTAMP(), nullable=True),
        sa.Column("validity_max", postgresql.TIMESTAMP(), nullable=True),
        sa.PrimaryKeyConstraint("id_idb_release"),
        sa.UniqueConstraint("idb_version", "idb_source"),
        schema="idb",
    )
    create_table(
        "palisade_metadata",
        sa.Column("id_palisade_metadata", sa.INTEGER(), nullable=False),
        sa.Column("palisade_version", sa.String(length=32), nullable=False),
        sa.Column("srdb_id", sa.String(length=10), nullable=True),
        sa.Column("palisade_id", sa.String(length=128), nullable=True),
        sa.Column("packet_category", sa.String(length=128), nullable=True),
        sa.PrimaryKeyConstraint("id_palisade_metadata"),
        sa.UniqueConstraint("palisade_version", "srdb_id"),
        schema="idb",
    )
    create_table(
        "item_info",
        sa.Column("id_item_info", sa.INTEGER(), nullable=False),
        sa.Column("idb_release_id", sa.INTEGER(), nullable=False),
        sa.Column("srdb_id", sa.String(length=10), nullable=True),
        sa.Column("palisade_id", sa.String(length=128), nullable=True),
        sa.Column("item_descr", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["idb_release_id"],
            ["idb.idb_release.id_idb_release"],
        ),
        sa.PrimaryKeyConstraint("id_item_info"),
        sa.UniqueConstraint("srdb_id", "idb_release_id"),
        schema="idb",
    )
    create_table(
        "packet_metadata",
        sa.Column("id_packet_metadata", sa.INTEGER(), nullable=False),
        sa.Column("idb_release_id", sa.INTEGER(), nullable=False),
        sa.Column("packet_category", sa.String(length=256), nullable=True),
        sa.ForeignKeyConstraint(
            ["idb_release_id"],
            ["idb.idb_release.id_idb_release"],
        ),
        sa.PrimaryKeyConstraint("id_packet_metadata"),
        schema="idb",
    )
    create_table(
        "packet_header",
        sa.Column("id_packet_header", sa.INTEGER(), nullable=False),
        sa.Column("item_info_id", sa.INTEGER(), nullable=False),
        sa.Column("packet_metadata_id", sa.INTEGER(), nullable=True),
        sa.Column("cat_eng", sa.String(length=128), nullable=False),
        sa.Column("cat", sa.BIGINT(), nullable=False),
        sa.Column("packet_type", sa.String(length=16), nullable=False),
        sa.Column("byte_size", sa.INTEGER(), nullable=False),
        sa.Column("pid_eng", sa.String(length=128), nullable=True),
        sa.Column("pid", sa.BIGINT(), nullable=True),
        sa.Column("spid", sa.BIGINT(), nullable=True),
        sa.Column("sid_eng", sa.String(length=128), nullable=True),
        sa.Column("sid", sa.BIGINT(), nullable=True),
        sa.Column("service_type_eng", sa.String(length=128), nullable=True),
        sa.Column("service_type", sa.BIGINT(), nullable=False),
        sa.Column("service_subtype_eng", sa.String(length=128), nullable=True),
        sa.Column("service_subtype", sa.BIGINT(), nullable=False),
        sa.ForeignKeyConstraint(
            ["item_info_id"],
            ["idb.item_info.id_item_info"],
        ),
        sa.ForeignKeyConstraint(
            ["packet_metadata_id"],
            ["idb.packet_metadata.id_packet_metadata"],
        ),
        sa.PrimaryKeyConstraint("id_packet_header"),
        schema="idb",
    )
    create_table(
        "param_info",
        sa.Column("id_param_info", sa.INTEGER(), nullable=False),
        sa.Column("item_info_id", sa.INTEGER(), nullable=False),
        sa.Column("par_bit_length", sa.SMALLINT(), nullable=False),
        sa.Column("par_type", sa.String(length=16), nullable=False),
        sa.Column("par_max", postgresql.DOUBLE_PRECISION(), nullable=True),
        sa.Column("par_min", postgresql.DOUBLE_PRECISION(), nullable=True),
        sa.Column("par_def", sa.String(length=64), nullable=True),
        sa.Column("cal_num", sa.String(length=16), nullable=True),
        sa.Column("cal_val", sa.BIGINT(), nullable=True),
        sa.Column("par_unit", sa.String(length=8), nullable=True),
        sa.Column("par_is_editable", sa.BOOLEAN(), nullable=True),
        sa.Column("transfer_function_id", sa.INTEGER(), nullable=True),
        sa.ForeignKeyConstraint(
            ["item_info_id"],
            ["idb.item_info.id_item_info"],
        ),
        sa.ForeignKeyConstraint(
            ["transfer_function_id"],
            ["idb.item_info.id_item_info"],
        ),
        sa.PrimaryKeyConstraint("id_param_info"),
        schema="idb",
    )
    create_table(
        "transfer_function",
        sa.Column("id_transfer_function", sa.INTEGER(), nullable=False),
        sa.Column("item_info_id", sa.INTEGER(), nullable=False),
        sa.Column("raw", sa.BIGINT(), nullable=False),
        sa.Column("eng", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["item_info_id"],
            ["idb.item_info.id_item_info"],
        ),
        sa.PrimaryKeyConstraint("id_transfer_function"),
        schema="idb",
    )
    create_table(
        "packet_map_info",
        sa.Column("id_packet_map_info", sa.INTEGER(), nullable=False),
        sa.Column("packet_header_id", sa.INTEGER(), nullable=False),
        sa.Column("param_info_id", sa.INTEGER(), nullable=False),
        sa.Column("block_size", sa.INTEGER(), nullable=False),
        sa.Column("group_size", sa.INTEGER(), nullable=False),
        sa.Column("loop_value", sa.INTEGER(), nullable=False),
        sa.Column("byte_position", sa.INTEGER(), nullable=False),
        sa.Column("bit_position", sa.INTEGER(), nullable=False),
        sa.ForeignKeyConstraint(
            ["packet_header_id"],
            ["idb.packet_header.id_packet_header"],
        ),
        sa.ForeignKeyConstraint(
            ["param_info_id"],
            ["idb.param_info.id_param_info"],
        ),
        sa.PrimaryKeyConstraint("id_packet_map_info"),
        schema="idb",
    )
    # ### end Alembic commands ###

    # added by hand
    # fix table permissions
    grant_user_seq = (
        "GRANT USAGE, SELECT ON ALL SEQUENCES IN " "SCHEMA idb TO {0}".format(user)
    )
    execute(grant_user_seq)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    drop_table("packet_map_info", schema="idb")
    drop_table("transfer_function", schema="idb")
    drop_table("param_info", schema="idb")
    drop_table("packet_header", schema="idb")
    drop_table("packet_metadata", schema="idb")
    drop_table("item_info", schema="idb")
    drop_table("palisade_metadata", schema="idb")
    drop_table("idb_release", schema="idb")

    # added by hand
    execute("DROP TYPE IF EXISTS idb.idb_source_type")
    execute("DROP SCHEMA idb")

    # ### end Alembic commands ###



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from roc.idb.parsers.idb_elements import Enumeration

__all__ = [
    "EnumerationParser",
]


class EnumerationParser:
    """
    Mixin class used to parse enumerations
    """

    def populate_enumerations(self):
        """
        Search in the root node enumeration definitions and populate it with
        information from the IDB.
        """
        # find the node for the definition of the enumeration
        nodes = self._findall(
            self.root,
            ".//{0}EnumDefinition",
            "",
        )

        # loop over nodes and create enumeration
        for node in nodes:
            palisade_id = node.attrib["ID"]

            for packet_type in ["TM", "TC"]:
                # get the srdb_id from the palisade_id and the packet type if any
                # if it's an internal enum (with no srdb id), replace the srdb_id by the palisade id
                srdb_id = self.palisade_to_srdb_id[packet_type].get(
                    palisade_id, palisade_id
                )

                if srdb_id is None:
                    continue

                enumeration = Enumeration(srdb_id)

                # get some information from the node
                enumeration.data_type = node.attrib["Type"]
                enumeration.values = {
                    value_entry.attrib["ID"]: int(value_entry.attrib["NumValue"], 0)
                    for value_entry in self._findall(node, ".//{0}Value", "")
                }
                # check for recursive enumeration
                enumeration.values = [
                    Enumeration(
                        self.palisade_to_srdb_id[packet_type].get(
                            enum_entry.attrib["IDRef"],
                            enum_entry.attrib["IDRef"],
                        )
                    )
                    for enum_entry in self._findall(node, ".//{0}Enum", "")
                ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os.path as osp
import xml.etree.ElementInclude as EI
import xml.etree.ElementTree as ET

from poppy.core.logger import logger
from roc.idb.parsers.idb_parser import IDBParser

__all__ = ["BasePALISADEParser", "PALISADEException"]


class PALISADEException(Exception):
    """
    Exception for PALISADE.
    """


class BasePALISADEParser(IDBParser):
    """
    Base class used for the parsing of the XML of the PALISADE IDB
    """

    def __init__(self, idb, mapping_file_path):
        """
        Store the file names of the XML files containing the IDB and the
        mapping to the SRDB.
        """
        super().__init__(mapping_file_path)
        self._source = "PALISADE"
        self.idb = osp.join(idb, "xml", "RPW_IDB.xml")
        self.enumerations = None

        # create the tree of the IDB
        self.store_tree()

        # store the palisade metadata as a map indexed by SRDB IDs
        self.palisade_metadata_map = {}

    def store_tree(self):
        """
        Create the tree from the file given in argument.
        """
        # check that the file exist
        if self.idb is None or not osp.isfile(self.idb):
            message = "IDB {0} doesn't exist".format(self.idb)
            logger.error(message)
            raise FileNotFoundError(message)

        # get the tree and namespace
        self.tree, self.namespace = self.parse_and_get_ns(self.idb)

        # get the root element
        self.root = self.tree.getroot()

        # get the directory path
        self._path = self.get_path()

        # include other files from root : First level
        EI.include(self.root, loader=self._loader)
        # include other files from child of root (second level)
        # FIXME : Implement the general case with several level of "include"
        EI.include(self.root, loader=self._loader)

        # compute the parent map after includes
        self.parent_map = {
            child: parent for parent in self.tree.iter() for child in parent
        }

    def get_path(self):
        """
        Return the directory path of the IDB to be used as root for the
        included files in the XML.
        """
        # get the absolute path of the directory
        return osp.abspath(osp.dirname(self.idb))

    def _loader(self, href, parse, encoding=None):
        """
        The loader of included files in XML. In our case, the parse attribute
        is always XML so do not do anything else.
        """
        # open the included file and parse it to be returned
        with open(osp.join(self._path, href), "rb") as f:
            return ET.parse(f).getroot()

    def _find(self, node, path, *args):
        """
        To find a node with the namespace passed as argument.
        """
        return node.find(
            path.format(*[self._ns[x] if x in self._ns else x for x in args])
        )

    def _findall(self, node, path, *args):
        """
        To findall a node with the namespace passed as argument.
        """
        return node.findall(
            path.format(*[self._ns[x] if x in self._ns else x for x in args])
        )

    def get_structures(self):
        """
        Get a dictionary of all structures defined in the XML IDB.
        """
        # get all structures
        structures = self._findall(
            self.root,
            ".//{0}StructDefinition",
            "",
        )
        # keep a dictionary of the structure name and the corresponding node
        self.structures = {struct.attrib["ID"]: struct for struct in structures}

    def generate_srdb_ids_from_palisade_id(self, palisade_id):
        """
        Given a PALISADE ID, generate the corresponding SRDB ID(s)
        """
        for packet_type in ["TM", "TC"]:
            if palisade_id in self.palisade_to_srdb_id[packet_type]:
                if isinstance(self.palisade_to_srdb_id[packet_type][palisade_id], list):
                    for srdb_id in self.palisade_to_srdb_id[packet_type][palisade_id]:
                        yield srdb_id
                else:
                    yield self.palisade_to_srdb_id[packet_type][palisade_id]

    def get_palisade_category(self, node):
        """
        Get the palisade category of a given node using the parent map of the xml tree

        :param node: the node we want the category
        :return: the palisade category
        """

        current_node = node

        node_tag_list = []

        while "parent is a node of the xml tree":
            # get the parent node
            parent = self.parent_map.get(current_node, None)

            if parent is None:
                break

            if parent.tag.endswith("Category"):
                # prepend the category to the list
                node_tag_list.insert(0, parent.attrib["Name"])

            current_node = parent

        return "/" + "/".join(node_tag_list)

    def is_enumeration(self, parameter):
        """
        Return True if a parameter is an enumeration.
        """
        return parameter.is_enumeration()

    @property
    def namespace(self):
        return self._ns

    @namespace.setter
    def namespace(self, namespace):
        self._ns = namespace

        # store some information from the namespace
        self.struct_tag = "{0}Struct".format(self._ns[""])
        self.param_tag = "{0}Parameter".format(self._ns[""])
        self.spare_tag = "{0}Spare".format(self._ns[""])
        self.loop_tag = "{0}Loop".format(self._ns[""])

    def version(self):
        """
        Return the version of the IDB defined in the XML file.
        """
        if "Version" not in self.root.attrib:
            message = "No version for the IDB from the specified XML file."
            logger.error(message)
            raise PALISADEException(message)
        return self.root.attrib["Version"].replace(" ", "_").upper()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Exception definition for roc.idb plugin.
"""

from poppy.core.logger import logger

__all__ = ["NoIdbFound", "MultipleIdbFound", "IdbQueryError", "IdbUpdateError"]


class NoIdbFound(Exception):
    """Exception raised if an IDB release is not found in the database."""

    def __init__(self, message, *args, **kwargs):
        super(NoIdbFound, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message

    pass


class MultipleIdbFound(Exception):
    """Exception raised if an IDB release is found more than one time in the database."""

    def __init__(self, message, *args, **kwargs):
        super(MultipleIdbFound, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message


class IdbQueryError(Exception):
    """Exception raised if an IDB query has failed."""

    def __init__(self, message, *args, **kwargs):
        super(IdbQueryError, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message


class IdbUpdateError(Exception):
    """Exception raised if an IDB update has failed."""

    def __init__(self, message, *args, **kwargs):
        super(IdbUpdateError, self).__init__(*args, **kwargs)
        logger.error(message)
        self.message = message


class SetTrangeIdbReleaseError(Exception):
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import csv
import datetime
import os

import xlwt
from poppy.core.target import PyObjectTarget
from poppy.core.task import Task
from roc.idb.parsers.mib_parser.procedure import Procedure

__all__ = ["ExportSequencesTask"]


def create_csv_from_dict(filepath, data):
    # create directories if needed
    os.makedirs(os.path.dirname(filepath), exist_ok=True)

    # create the csv file
    with open(filepath, "w") as csv_file:
        csv_writer = csv.DictWriter(csv_file, data["header"])
        csv_writer.writeheader()
        for row in data["rows"]:
            csv_writer.writerow(row)


def create_sequence_csv(filepath, data):
    # loop over the sheets
    for key in ["STMT", "CMD Params", "TLM Values", "PKT Params", "Info"]:
        create_csv_from_dict(os.path.join(filepath, key + ".csv"), data[key])


def create_xls_sheet(workbook, sheet_name, data):
    header = data["header"]

    # create the sheet
    sheet = workbook.add_sheet(sheet_name)
    # populate the header
    for index, entry in enumerate(header):
        sheet.write(0, index, entry)

    # populate the rest of the sheet with the given data
    for row_idx, row in enumerate(data["rows"]):
        for col_index, col_name in enumerate(header):
            entry = row.get(col_name)

            # handle case where style is bundled with the value
            if isinstance(entry, tuple):
                value, style = entry
                sheet.write(row_idx + 1, col_index, value, style)
            else:
                sheet.write(row_idx + 1, col_index, entry)


def create_sequence_xls(filepath, data):
    workbook = xlwt.Workbook()

    # loop over the sheets
    for key in ["STMT", "CMD Params", "TLM Values", "PKT Params", "Info"]:
        create_xls_sheet(workbook, key, data[key])

    workbook.save(filepath)


def render_xls_time(time_str, format="h:mm:ss"):
    h, m, s = map(int, time_str.split(":"))
    time_format = xlwt.XFStyle()
    time_format.num_format_str = format
    return (datetime.time(h, m, s), time_format)


@PyObjectTarget.input("sequences")
@PyObjectTarget.input("tc_duration_table")
@Task.as_task(plugin_name="roc.idb", name="idb_export_sequences")
def ExportSequencesTask(self):
    args = self.pipeline.args

    if args.seq_dir_path is None:
        output_path = os.path.join(self.pipeline.output, "mib_sequences")
    else:
        output_path = args.seq_dir_path

    sequences_data = self.inputs["sequences"].value
    tc_duration_table = self.inputs["tc_duration_table"].value

    # create the csv files
    for procedure_name in sequences_data:
        for sequence_name in sequences_data[procedure_name]:
            xls_data = Procedure.as_xls_data(
                sequences_data[procedure_name][sequence_name],
                tc_duration_table,
            )
            create_sequence_csv(
                os.path.join(output_path, procedure_name, sequence_name),
                xls_data,
            )

            # update the total duration value to be compatible with the xls renderer
            xls_data["Info"]["rows"][0]["Procedure Duration"] = render_xls_time(
                xls_data["Info"]["rows"][0]["Procedure Duration"]
            )

            create_sequence_xls(
                os.path.join(output_path, procedure_name, sequence_name + ".xls"),
                xls_data,
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.generic.metaclasses import Singleton
from poppy.core.logger import logger

__all__ = ["IDBToNumpy"]


class IDBToNumpy(object, metaclass=Singleton):
    """
    A class to convert IDB types to the good numpy type.
    """

    def __init__(self):
        self.mapping = {
            "b1Bool": "uint8",
            "bU1": "uint8",
            "bU2": "uint8",
            "bU3": "uint8",
            "bU4": "uint8",
            "bU5": "uint8",
            "bU6": "uint8",
            "bU7": "uint8",
            "bU8": "uint8",
            "bU9": "uint16",
            "bU10": "uint16",
            "bU12": "uint16",
            "bU24": "uint32",
            "UShort": "uint16",
            "Short": "int16",
            "b10": "int16",
            "UInt": "uint32",
            "Float": "float32",
        }

    def __call__(self, value):
        if value not in self.mapping:
            logger.error("{0} type not mapped to C type".format(value))
            return
        return self.mapping[value]


# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import shutil

import pytest

from poppy.core.logger import logger
from poppy.core.test import CommandTestCase

from roc.idb.models.idb import PacketHeader, ItemInfo, IdbRelease
from .idb_test_util import reset_db, load_idb


class TestLoadIdbCommand(CommandTestCase):
    @pytest.mark.parametrize(
        "idb_source,idb_version",
        [("PALISADE", "4.3.5_MEB_PFM"), ("MIB", "20200131")],
    )
    def test_install_and_load_idb(self, idb_source, idb_version):
        # Initializing database
        logger.debug("Reset database ...")
        reset_db(self)

        # Installing IDB
        logger.debug(f"Installing IDB [{idb_source}-{idb_version}] ...")
        self.install_dir = load_idb(self, idb_source, idb_version)

        logger.debug(f"Querying IDB [{idb_source}-{idb_version}] ...")
        packet_header = (
            self.session.query(PacketHeader)
            .join(ItemInfo)
            .join(IdbRelease)
            .filter(
                ItemInfo.srdb_id == "YIW00083",
                IdbRelease.idb_version == idb_version,
                IdbRelease.idb_source == idb_source,
            )
            .one()
        )

        # make assertions
        assert packet_header.cat == 7
        assert packet_header.sid == 42122

    def teardown_method(self, method):
        """
        Method called immediately after the test method has been called and the result recorded.

        This is called even if the test method raised an exception.

        :param method: the test method
        :return:
        """

        # rollback the database
        super().teardown_method(method)

        # Remove IDB folder
        shutil.rmtree(self.install_dir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.generic.metaclasses import Singleton

__all__ = ["IDBToBit"]


class IDBToBitMappingException(Exception):
    pass


class IDBToBit(object, metaclass=Singleton):
    """
    A class to convert IDB types to the good bit length.
    """

    def __init__(self):
        self.mapping = {
            "b1Bool": 1,
            "bU1": 1,
            "bU2": 2,
            "bU3": 3,
            "bU4": 4,
            "bU5": 5,
            "bU6": 6,
            "bU7": 7,
            "bU8": 8,
            "bU9": 9,
            "bU10": 10,
            "bU12": 12,
            "bU24": 24,
            "UShort": 16,
            "b10": 10,
            "Short": 16,
            "UInt": 32,
            "Float": 32,
        }

    def __call__(self, value):
        if value not in self.mapping:
            raise IDBToBitMappingException("{0} type not mapped".format(value))
        return self.mapping[value]


# vim: set tw=79 :



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.db.base import Base
from poppy.core.db.non_null_column import NonNullColumn

from sqlalchemy import String, UniqueConstraint
from sqlalchemy.dialects.postgresql import INTEGER
from sqlalchemy.orm import validates

__all__ = [
    "PalisadeMetadata",
]


class PalisadeMetadata(Base):
    """
    The "palisade_metadata" table provides PALISADE metadata for each IDB packet/parameter

    This table is mainly used to store the mapping between PALISADE and SRDB IDs.
    Moreover, it contains the categories associated with each packet
    """

    id_palisade_metadata = NonNullColumn(INTEGER(), primary_key=True)

    palisade_version = NonNullColumn(String(32), descr="Version of the PALISADE IDB")

    srdb_id = NonNullColumn(
        String(10), nullable=True, descr="SRDB ID of the packet or parameter"
    )

    palisade_id = NonNullColumn(
        String(128),
        nullable=True,
        descr="PALISADE ID of the packet or parameter",
    )

    packet_category = NonNullColumn(
        String(128),
        nullable=True,
        descr="The packet category (null for parameters)",
    )

    __tablename__ = "palisade_metadata"
    __table_args__ = (
        UniqueConstraint("palisade_version", "srdb_id"),
        {"schema": "idb"},
    )

    @validates("srdb_id")
    def validate_srdbid(self, key, srdb_id):
        if srdb_id is None:
            pass
        elif srdb_id.startswith("CIW"):
            assert len(srdb_id) == 10
        else:
            assert len(srdb_id) == 8
        return srdb_id



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.db.database import Database
from poppy.core.conf import settings

__all__ = ["Base"]

# register the base for future use
Base = Database.bases_manager.get(settings.MUSIC_DATABASE)



# -*- coding: utf-8 -*-
from poppy.core.command import Command
from roc.idb.tasks import ListReleasesTask, ListPalisadeMetadataTask

__all__ = ["List", "ListReleases", "ListPalisadeMetadata"]


class List(Command):
    """
    Command to list the IDB content
    """

    __command__ = "idb_list"
    __command_name__ = "list"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "List the IDB content"


class ListReleases(Command):
    """
    Command to list the different idb releases
    """

    __command__ = "idb_list_releases"
    __command_name__ = "releases"
    __parent__ = "idb_list"
    __parent_arguments__ = ["base"]
    __help__ = "Command to list the different idb releases"

    def setup_tasks(self, pipeline):
        """
        List the different idb releases
        """

        # the task
        list_idb_releases = ListReleasesTask()

        # set the pipeline for this situation
        pipeline | list_idb_releases
        pipeline.start = list_idb_releases


class ListPalisadeMetadata(Command):
    """
    Command to list the different idb releases
    """

    __command__ = "idb_list_palisade_metadata"
    __command_name__ = "palisade_metadata"
    __parent__ = "idb_list"
    __parent_arguments__ = ["base"]
    __help__ = "Command to list the palisade metadata"

    def setup_tasks(self, pipeline):
        """
        List the palisade metadata
        """

        # the task
        list_palisade_metadata = ListPalisadeMetadataTask()

        # set the pipeline for this situation
        pipeline | list_palisade_metadata
        pipeline.start = list_palisade_metadata


# TODO: List packets/parameters



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Reference documentation
# ROC-GEN-SYS-NTT-00038-LES_Iss01_Rev02(Mission_Database_Description_Document)

from poppy.core.db.base import Base
from poppy.core.db.non_null_column import NonNullColumn

from sqlalchemy import String, ForeignKey, UniqueConstraint
from sqlalchemy.dialects.postgresql import (
    BIGINT,
    BOOLEAN,
    DOUBLE_PRECISION,
    ENUM,
    INTEGER,
    SMALLINT,
    TIMESTAMP,
)
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relationship, validates, backref

__all__ = [
    "ItemInfo",
    "PacketMapInfo",
    "IdbRelease",
    "PacketMetadata",
    "ParamInfo",
    "PacketHeader",
    "TransferFunction",
]

idb_source_enum = ["PALISADE", "MIB", "SRDB"]


class IdbRelease(Base):
    """
    The “idb_release” table provides information about the IDB releases stored
    in the database.
    """

    id_idb_release = NonNullColumn(INTEGER(), primary_key=True)

    idb_version = NonNullColumn(String(128), descr="Version of the RPW IDB")

    idb_source = NonNullColumn(
        ENUM(*idb_source_enum, name="idb_source_type", schema="idb"),
        descr="Sources of the IDB. Possible values " f"are : {idb_source_enum}",
    )

    mib_version = NonNullColumn(
        String(),
        nullable=True,
        descr="First row of the vdf.dat table including tabs",
    )

    current = NonNullColumn(
        BOOLEAN(),
        descr='"True" = current IDB to be used' ' by default, "False" otherwise',
    )

    validity_min = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr="Minimal date/time of the IDB validity",
    )

    validity_max = NonNullColumn(
        TIMESTAMP(),
        nullable=True,
        descr="Maximal date/time of the IDB validity",
    )

    __tablename__ = "idb_release"
    __table_args__ = (
        UniqueConstraint("idb_version", "idb_source"),
        {
            "schema": "idb",
        },
    )


class ItemInfo(Base):
    """
    The “item_info” table provides general information about RPW TM/TC packets
    and packet parameters defined in the IDB.
    """

    id_item_info = NonNullColumn(INTEGER(), primary_key=True)

    idb_release_id = NonNullColumn(
        INTEGER(), ForeignKey("idb.idb_release.id_idb_release")
    )

    srdb_id = NonNullColumn(
        String(10), nullable=True, descr="SRDB ID of the packet or parameter"
    )

    palisade_id = NonNullColumn(
        String(128),
        nullable=True,
        descr="PALISADE ID of the packet or parameter",
    )

    item_descr = NonNullColumn(
        String(), nullable=True, descr="Description of the packet or parameter"
    )

    idb = relationship(
        "IdbRelease",
        backref=backref("item_info_idb_release", cascade="all,delete-orphan"),
    )

    __tablename__ = "item_info"
    __table_args__ = (
        UniqueConstraint("srdb_id", "idb_release_id"),
        {
            "schema": "idb",
        },
    )

    @validates("srdb_id")
    def validate_srdbid(self, key, srdb_id):
        if srdb_id is None:
            pass
        elif srdb_id.startswith("CIW"):
            assert len(srdb_id) == 10
        else:
            assert len(srdb_id) == 8
        return srdb_id


class PacketMetadata(Base):
    """
    The “packet_metadata” table provides information about the SCOS-2000 packet
    description.
    """

    id_packet_metadata = NonNullColumn(INTEGER(), primary_key=True)

    idb_release_id = NonNullColumn(
        INTEGER(), ForeignKey("idb.idb_release.id_idb_release")
    )

    packet_category = NonNullColumn(String(256), nullable=True, descr="Packet category")

    idb = relationship(
        "IdbRelease",
        backref=backref("packet_metadata_idb_release", cascade="all,delete-orphan"),
    )

    __tablename__ = "packet_metadata"
    __table_args__ = {
        "schema": "idb",
    }


class TransferFunction(Base):
    """
    The "transfer_function" table provides information about the transfer functions.
    """

    id_transfer_function = NonNullColumn(INTEGER(), primary_key=True)

    item_info_id = NonNullColumn(INTEGER(), ForeignKey("idb.item_info.id_item_info"))

    raw = NonNullColumn(BIGINT(), descr="Raw value of the parameter")

    eng = NonNullColumn(
        String(),
        descr="Engineering value of the parameter after transfer function applies",
    )

    item_info = relationship(
        "ItemInfo",
        backref=backref("transfer_function_item_info", cascade="all,delete-orphan"),
    )

    __tablename__ = "transfer_function"
    __table_args__ = {
        "schema": "idb",
    }


class PacketHeader(Base):
    """
    The “packet_header” table provides information about the CCSDS packet
    header description.
    """

    id_packet_header = NonNullColumn(INTEGER(), primary_key=True)
    item_info_id = NonNullColumn(INTEGER(), ForeignKey("idb.item_info.id_item_info"))
    packet_metadata_id = NonNullColumn(
        INTEGER(),
        ForeignKey("idb.packet_metadata.id_packet_metadata"),
        nullable=True,
    )
    cat_eng = NonNullColumn(String(128), descr="Packet category in engineering value")
    cat = NonNullColumn(BIGINT(), descr="Packet category in raw value")
    packet_type = NonNullColumn(String(16), descr="Packet type in engineering value")
    byte_size = NonNullColumn(INTEGER(), descr="Packet byte size")
    pid_eng = NonNullColumn(
        String(128), nullable=True, descr="Packet PID in engineering value"
    )
    pid = NonNullColumn(BIGINT(), nullable=True, descr="Packet PID in raw value")
    spid = NonNullColumn(BIGINT(), nullable=True, descr="Packet SPID")
    sid_eng = NonNullColumn(
        String(128), nullable=True, descr="Packet SID in engineering value"
    )
    sid = NonNullColumn(BIGINT(), nullable=True, descr="Packet SID in raw value")
    service_type_eng = NonNullColumn(
        String(128),
        nullable=True,
        descr="Packet service type in engineering" " value",
    )
    service_type = NonNullColumn(BIGINT(), descr="Packet service type in raw value")
    service_subtype_eng = NonNullColumn(
        String(128),
        nullable=True,
        descr="Packet service subtype in " "engineering value",
    )
    service_subtype = NonNullColumn(
        BIGINT(), descr="Packet service subtype in raw value"
    )

    item_info = relationship(
        "ItemInfo",
        backref=backref("packet_header_item_info", cascade="all,delete-orphan"),
    )
    packet_metadata = relationship(
        "PacketMetadata",
        backref=backref("packet_header_packet_metadata", cascade="all,delete-orphan"),
    )

    __tablename__ = "packet_header"
    __table_args__ = {
        "schema": "idb",
    }


class ParamInfo(Base):
    """
    The “param_info” table provides information about the CCSDS packet
    parameter description.
    """

    id_param_info = NonNullColumn(INTEGER(), primary_key=True)
    item_info_id = NonNullColumn(INTEGER(), ForeignKey("idb.item_info.id_item_info"))
    par_bit_length = NonNullColumn(SMALLINT(), descr="Bit length of the parameter")
    par_type = NonNullColumn(String(16), descr="Type of parameter")
    par_max = NonNullColumn(
        DOUBLE_PRECISION(), nullable=True, descr="Parameter maximum value"
    )
    par_min = NonNullColumn(
        DOUBLE_PRECISION(), nullable=True, descr="Parameter minimum value"
    )
    par_def = NonNullColumn(String(64), nullable=True, descr="Parameter default value")
    cal_num = NonNullColumn(String(16), nullable=True, descr="raw-eng. cal. num.")
    cal_val = NonNullColumn(BIGINT(), nullable=True, descr="raw-eng. cal. val.")
    par_unit = NonNullColumn(String(8), nullable=True, descr="Parameter unit")
    par_is_editable = NonNullColumn(
        BOOLEAN, nullable=True, descr="Parameter edition possibility"
    )

    transfer_function_id = NonNullColumn(
        INTEGER(), ForeignKey("idb.item_info.id_item_info"), nullable=True
    )

    __tablename__ = "param_info"
    __table_args__ = {
        "schema": "idb",
    }

    info = relationship(
        ItemInfo,
        foreign_keys=[item_info_id],
        backref=backref("param_info_item_info", cascade="all,delete-orphan"),
    )
    transfer_function = relationship(
        ItemInfo,
        foreign_keys=[transfer_function_id],
        backref=backref("param_info_transfer_function", cascade="all,delete-orphan"),
    )

    name = association_proxy("info", "palisade_id")
    description = association_proxy("info", "item_descr")


class PacketMapInfo(Base):
    """
    The “packet_map_info” table provides mapping information between the packet
    and parameter description.
    """

    id_packet_map_info = NonNullColumn(INTEGER(), primary_key=True)
    packet_header_id = NonNullColumn(
        INTEGER(), ForeignKey("idb.packet_header.id_packet_header")
    )
    param_info_id = NonNullColumn(INTEGER(), ForeignKey("idb.param_info.id_param_info"))
    block_size = NonNullColumn(INTEGER(), descr="Size of the block")
    group_size = NonNullColumn(INTEGER(), descr="Group size")
    loop_value = NonNullColumn(INTEGER(), descr="Number of loop over blocks")
    byte_position = NonNullColumn(INTEGER(), descr="Byte position")
    bit_position = NonNullColumn(INTEGER(), descr="Bit position")

    __tablename__ = "packet_map_info"
    __table_args__ = {
        "schema": "idb",
    }

    packet = relationship(
        "PacketHeader",
        backref=backref("packet_parameters", cascade="all,delete-orphan"),
    )
    parameter = relationship(
        "ParamInfo",
        backref=backref("packet_map_info_param_info", cascade="all,delete-orphan"),
    )

    def __init__(
        self,
        parameter,
        packet,
        byte_position=0,
        bit_position=0,
        group_size=0,
        block_size=0,
        loop_value=0,
    ):
        self.parameter = parameter
        self.packet = packet
        self.byte_position = byte_position
        self.bit_position = bit_position
        self.group_size = group_size
        self.block_size = block_size
        self.loop_value = loop_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from datetime import datetime

from poppy.core.logger import logger

from roc.idb.constants import MIB_VERSION_STRFORMAT

__all__ = ["valid_mib_version"]


def valid_mib_version(mib_version, valid_format=MIB_VERSION_STRFORMAT):
    """
    Check if input MIB version is valid.

    :param mib_version: String containing the MIB version to check
    :param valid_format: string containing the valid MIB version format
    (use datetime.strptime convention)
    :return: mib version string if valid, otherwise raise an ValueError exception
    """

    try:
        # Use datetime to check if MIB version is valid
        _ = datetime.strptime(mib_version, valid_format)
    except ValueError:
        logger.error(f"Input MIB version {mib_version} is not valid!")
        raise

    return mib_version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os.path

from poppy.core.conf import settings
from poppy.core.db.connector import Connector
from poppy.core.db.handlers import get_or_create
from poppy.core.logger import logger
from poppy.core.task import Task
from roc.idb.models.palisade_metadata import PalisadeMetadata
from roc.idb.parsers import PALISADEMetadataParser

__all__ = ["LoadPalisadeMetadataTask", "put_palisade_metadata"]


def put_palisade_metadata(session, parser):
    """
    Put PALISADE metadata into roc database
    """

    # get the palisade version
    palisade_version = parser.version()

    for srdb_id in parser.palisade_metadata_map:
        palisade_id = parser.palisade_metadata_map[srdb_id]["palisade_id"]

        get_or_create(
            session,
            PalisadeMetadata,
            palisade_version=palisade_version,
            srdb_id=srdb_id.strip(),
            palisade_id=palisade_id.strip()
            if isinstance(palisade_id, str)
            else palisade_id,
            packet_category=parser.palisade_metadata_map[srdb_id][
                "palisade_category"
            ].strip(),
        )


class LoadPalisadeMetadataTask(Task):
    plugin_name = "roc.idb"

    name = "idb_load_palisade_metadata"

    @Connector.if_connected(settings.PIPELINE_DATABASE)
    def run(self):
        """
        Parse the PALISADE XML database to insert PALISADE metadata into the ROC
        database.
        """
        # get the database session
        session = self.pipeline.db.session

        idb_path = self.pipeline.args.idb_path

        mapping_path = os.path.join(
            idb_path, "xml", "IDB_SRDB_Mapping_Table_by_MapGen.xml"
        )

        # create the metadata parser
        parser = PALISADEMetadataParser(idb_path, mapping_path)

        # parse the XML file
        logger.debug("Parsing the Palisade metadata")
        parser.parse()

        # put palisade metadata into the database
        put_palisade_metadata(session, parser)

        # update the database
        self.pipeline.db.update_database()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Test load_idb_dump command
"""

import shutil
import os

from sqlalchemy.orm.exc import NoResultFound
import pytest

from poppy.core.logger import logger
from poppy.core.test import CommandTestCase

from roc.idb.constants import IDB_CACHE_DIR
from roc.idb.models.idb import IdbRelease
from .idb_test_util import reset_db


class TestLoadIdbDumpCommand(CommandTestCase):
    @pytest.mark.skip("Too long test: to be reworked")
    @pytest.mark.parametrize(
        "idb_source,idb_version,idb_dump_file,user,password",
        [
            (
                "MIB",
                "20200131",
                "https://rpw.lesia.obspm.fr/roc/data/private/devtest/roc/test_data/idb_dump/MIB_20200131/idb_dump.sql",
                os.environ.get("ROC_TEST_USERNAME", "roctest"),
                os.environ.get("ROC_TEST_PASSWORD", None),
            )
        ],
    )
    def test_load_idb_dump(
        self, idb_source, idb_version, idb_dump_file, user, password
    ):
        # Build idb_dump local path
        self.install_dir = os.path.join(
            IDB_CACHE_DIR, f"idb-{idb_source}-{idb_version}"
        )
        self.local_idb_dump = os.path.join(
            self.install_dir, os.path.basename(idb_dump_file)
        )
        os.makedirs(self.install_dir, exist_ok=True)

        if not password:
            logger.warning(f"Password is not defined for {user}!")

        # Reset database
        reset_db(self)

        # Loading IDB dump file
        command_to_test = [
            "pop",
            "-ll",
            "INFO",
            "idb",
            "--force",
            "load_idb_dump",
            "-i",
            self.install_dir,
            "-d",
            idb_dump_file,
            "-a",
            user,
            password,
        ]

        # Make sure the idb was not already loaded
        logger.debug(f'Running command: {" ".join(command_to_test)} ...')
        self.run_command(command_to_test)

        # Check idb sql dump file exits
        assert os.path.isfile(self.local_idb_dump)

        # Verify expected behaviour
        try:
            _ = (
                self.session.query(IdbRelease)
                .filter(
                    IdbRelease.idb_version == idb_version,
                    IdbRelease.idb_source == idb_source,
                )
                .one()
            )
        except NoResultFound:
            assert False
        else:
            assert True

    def teardown_method(self, method):
        """
        Method called immediately after the test method has been called and the result recorded.

        This is called even if the test method raised an exception.

        :param method: the test method
        :return:
        """

        # rollback the database
        super().teardown_method(method)

        # Remove IDB folder
        shutil.rmtree(self.install_dir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os

from poppy.core.test import CommandTestCase

from roc.idb.constants import IDB_CACHE_DIR


def reset_db(command_test_case: CommandTestCase, log_level: str = "INFO"):
    # Clear database
    clear_db_cmd = ["pop", "db", "downgrade", "base", "-ll", log_level]
    command_test_case.run_command(clear_db_cmd)

    # Run database migrations
    create_db_cmd = ["pop", "db", "upgrade", "heads", "-ll", log_level]
    command_test_case.run_command(create_db_cmd)


def load_idb(
    command_test_case: CommandTestCase,
    idb_source: str,
    idb_version: str,
    install_dir: str = None,
    log_level="INFO",
) -> str:
    if install_dir is None:
        install_dir = os.path.join(IDB_CACHE_DIR, f"idb-{idb_source}-{idb_version}")
        os.makedirs(install_dir, exist_ok=True)

    # IDB loading
    load_idb_cmd = [
        "pop",
        "idb",
        "install",
        "--force",
        "--install-dir",
        install_dir,
        "-s",
        idb_source,
        "-v",
        idb_version,
        "--load",
        "-ll",
        log_level,
    ]
    command_test_case.run_command(load_idb_cmd)

    return install_dir



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__all__ = [
    "TFParser",
]


class TFParser:
    """
    Mixin class used to parse transfer functions
    """

    def create_tf(self):
        """
        Extract and create packets with TF (PALISADE_ID, Raw, Eng)
        """

        self.tf_dict = {}

        palisade_tf_dict = {}

        # loop over nodes
        for child in self._findall(
            self.root,
            ".//{0}CalibrationDefinition",
            "",
        ):
            palisade_id = child.attrib["ID"]
            palisade_tf_dict.setdefault(palisade_id, [])
            for subchild in self._findall(
                child,
                ".//{0}Tab",
                "",
            ):
                palisade_tf_dict[palisade_id].append(
                    {"raw": subchild.attrib["X"], "eng": subchild.attrib["Y"]}
                )

        # loop over the srdb_ids and retrieve the node info
        for palisade_id in palisade_tf_dict:
            for srdb_id in self.generate_srdb_ids_from_palisade_id(palisade_id):
                self.tf_dict[srdb_id] = palisade_tf_dict[palisade_id]



# -*- coding: utf-8 -*-
from poppy.core.command import Command
from roc.idb.tasks import ParseSequencesTask, ExportSequencesTask

__all__ = ["Export", "ExportSequences"]


class Export(Command):
    """
    Command to set a given idb version as current
    """

    __command__ = "idb_export"
    __command_name__ = "export"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "Export the IDB in the requested format"


class ExportSequences(Command):
    """
    Command to set a given idb version as current
    """

    __command__ = "idb_export_sequences"
    __command_name__ = "sequences"
    __parent__ = "idb_export"
    __parent_arguments__ = ["base"]
    __help__ = "Command to export CSV/XLS sequences using MIB files or the database"

    def add_arguments(self, parser):
        # path to MIB files directory
        parser.add_argument(
            "-mib",
            "--mib-dir-path",
            help="""
            Path to MIB files directory.
            """,
            type=str,
            required=True,
        )

        # path to sequences output path
        parser.add_argument(
            "-seq",
            "--seq-dir-path",
            help="""
            Path where sequences will be written.
            """,
            type=str,
        )

    def setup_tasks(self, pipeline):
        """
        Export MIB sequences as XLS/CSV
        """
        # instantiate the task
        parse_sequences = ParseSequencesTask()
        export_sequences = ExportSequencesTask()

        # set the pipeline topology
        pipeline | (parse_sequences | export_sequences)
        pipeline.start = parse_sequences



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math

import numpy
from poppy.core.generic.cache import CachedProperty
from poppy.core.logger import logger
from poppy.core.db.connector import Connector
from roc.idb.models.idb import (
    PacketHeader,
    ParamInfo,
    ItemInfo,
    IdbRelease,
    PacketMapInfo,
    TransferFunction,
)
from sqlalchemy.orm import aliased
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound

__all__ = ["IDBManager", "IDB", "PacketData"]


class IdbManagerError(Exception):
    pass


class PacketData(object):
    """
    A class for storing information and meta data about packets from the IDB.
    """

    def __init__(
        self,
        name,
        byte_size,
        description,
        parameters,
        srdbid,
        spid,
        pid,
        stype,
        subtype,
        sid,
        cat,
    ):
        self.name = name  # palisadeid
        self.srdbid = srdbid
        self.spid = spid
        self.pid = pid
        self.type = stype
        self.subtype = subtype
        self.sid = sid
        self.byte_size = byte_size
        self.description = description
        self.parameters = parameters
        self.category = cat


class IDB(object):
    """
    Class used as a container for the information about packets and parameters
    from the IDB database, with versioning.
    """

    def __init__(self, source, version, session):
        self.source = source
        self.version = version
        self.session = session

    @CachedProperty
    def packets(self):
        """
        Store in a cached property the structure of the defined packets in the
        IBD into a dictionary, whose keys are packets name and the values an
        instance of the :class:`~PacketData` structure.
        """
        from roc.rpl.packet_structure.parameter import Parameter

        # get the version of the IDB to use
        try:
            idb = self.get_idb()
        except NoResultFound:
            raise IdbManagerError(
                f"IDB ({self.source}, {self.version}) was not found in the database"
            )

        # make the query to load all necessary information
        info_alias = aliased(ItemInfo, name="param")
        query = self.session.query(
            ItemInfo.palisade_id,
            info_alias.palisade_id,
            PacketMapInfo.byte_position,
            PacketMapInfo.bit_position,
            PacketMapInfo.group_size,
            PacketMapInfo.block_size,
            PacketMapInfo.loop_value,
            ParamInfo.par_bit_length,
            ParamInfo.par_type,
            ParamInfo.par_max,
        )
        query = query.filter_by(idb=idb)
        query = query.join(
            PacketHeader, ItemInfo.id_item_info == PacketHeader.item_info_id
        )
        query = query.join(PacketMapInfo)
        query = query.join(ParamInfo)
        query = query.join(
            info_alias, info_alias.id_item_info == ParamInfo.item_info_id
        )
        query = query.order_by(
            ItemInfo.palisade_id,
            PacketMapInfo.byte_position,
            PacketMapInfo.bit_position,
        )

        # query data
        data = query.all()

        # query names of packets and their meta data
        query = self.session.query(
            ItemInfo.palisade_id,
            ItemInfo.item_descr,
            PacketHeader.byte_size,
            ItemInfo.srdb_id,
            PacketHeader.spid,
            PacketHeader.pid,
            PacketHeader.service_type,
            PacketHeader.service_subtype,
            PacketHeader.sid,
            PacketHeader.cat,
        )
        query = query.filter_by(idb=idb)
        query = query.join(PacketHeader)
        names = query.all()

        # now we create a map for packets and their list of parameters ordered
        # by byte and bit position, to have the order to work with, and also
        # other information on the data of the parameters.
        packets = {}

        # populate the dictionary of packets
        for (
            name,
            description,
            size,
            srdbid,
            spid,
            pid,
            stype,
            subtype,
            sid,
            cat,
        ) in names:
            packets[name] = PacketData(
                name=name,
                byte_size=size,
                description=description,
                parameters=[],
                srdbid=srdbid,
                spid=spid,
                pid=pid,
                stype=stype,
                subtype=subtype,
                sid=sid,
                cat=cat,
            )

        # now loop over parameters to associate them to the good packet
        for values in data:
            # transform none maximal value to zero
            val = [x for x in values]
            if val[-1] is None:
                val[-1] = 0
            if math.isnan(val[-1]) is True:
                val[-1] = numpy.nan

            logger.debug(f"Creation of parameter {val}")
            packets[values[0]].parameters.append(Parameter(*val[1:]))
        return packets

    @CachedProperty
    def parameters(self):
        # get the version of the IDB to use
        idb = self.get_idb()

        # same for parameters
        query = self.session.query(
            ItemInfo.palisade_id,
            ParamInfo,
        )
        query = query.filter_by(idb=idb)
        query = query.join(ParamInfo)

        # store packet names in a set for fast search
        parameters = query.all()
        return {x[0]: x[1] for x in parameters}

    @CachedProperty
    def transfer_function(self):
        """
        Cached Property for transfer function

        :return: dictionary with transfer function SRDB ID as keyword and list of ('raw', 'eng') as value
        """

        # get the version of the IDB to use
        idb = self.get_idb()

        # same for parameters
        query = self.session.query(
            ItemInfo.srdb_id,
            TransferFunction.raw,
            TransferFunction.eng,
        )
        query = query.filter_by(idb=idb)
        query = query.join(TransferFunction)

        # Sort by increasing raw values
        query = query.order_by(TransferFunction.raw)

        # Get transfer function
        tf_entries = query.all()

        # Initialize output dictionary containing set of raw, eng values of each TF SRDB ID
        tf_dict = {}
        for x in tf_entries:
            if x.srdb_id not in tf_dict:
                tf_dict[x.srdb_id] = []
            tf_dict[x.srdb_id].append((x.raw, x.eng))

        return tf_dict

    def get_idb(self):
        logger.debug(f"Getting IDB ({self.source}, {self.version})")
        query = self.session.query(IdbRelease)
        query = query.filter_by(idb_source=self.source, idb_version=self.version)
        return query.one()

    @staticmethod
    @Connector.if_connected("MAIN-DB")
    def get_version(current_datetime, idb_source="MIB", session=None):
        """
        Get the version of the working IDB for a given time.

        :param current_datetime: Time of the working IDB version to retrieve
        :param idb_source: IDB Source ('MIB', 'SRDB' or 'PALISADE')
        :param session: database session
        :return: string containing the version of the IDB
        """

        if session is None:
            # get a database session
            session = Connector.manager["MAIN-DB"].session

        # Build filters of query
        filters = [IdbRelease.idb_source == idb_source]
        filters.append(current_datetime >= IdbRelease.validity_min)
        filters.append(current_datetime < IdbRelease.validity_max)

        result = []
        try:
            query = session.query(IdbRelease)
            query = query.filters(**filters)
            result = query.one()
        except NoResultFound:
            logger.info(
                f"No IDB version found for {str(current_datetime)} "
                f'and source "{idb_source}"'
            )
        except MultipleResultsFound:
            logger.warning(
                f"Multiple IDB versions found for {str(current_datetime)} "
                f'and source "{idb_source}"!'
            )
        except Exception as e:
            logger.error(f"querying {query} has failed!")
            logger.debug(e)

        return result


class IDBManager(object):
    """
    A manager for the IDB interface acting like a cache for several IDB
    versions.
    """

    def __init__(self, session):
        """
        Provide a session for the connection with the database when requesting
        information about an IDB version.
        """
        self.session = session

        # init the cache
        self._cache = {}

    def __getitem__(self, source_version_tuple):
        """
        Create the IDB interface if not already present in the cache. If the
        version in argument is in the cache, returns the correct IDB.
        """
        # already loaded, provide information
        if source_version_tuple in self._cache:
            return self._cache[source_version_tuple]

        idb_source, idb_version = source_version_tuple

        idb_source_list = ["MIB", "SRDB", "PALISADE"]

        if idb_source not in idb_source_list:
            raise Exception(
                f"The idb source shall be one of the following item: {idb_source_list}"
            )

        # create an IDB and add it to the cache
        idb = IDB(
            idb_source,
            idb_version,
            self.session,
        )
        self._cache[source_version_tuple] = idb
        return idb

    def delete(self, version):
        """
        Delete the version in argument from the manager (the cache) if present.
        """
        if version in self._cache:
            self._cache.pop(version)

    def clear(self):
        """
        Clear the cache of the manager, in case you need to reload things with
        a new state.
        """
        self._cache.clear()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy
from sqlalchemy.orm.exc import NoResultFound

from poppy.core.conf import settings
from poppy.core.db.connector import Connector
from poppy.core.db.handlers import get_or_create
from poppy.core.logger import logger
from poppy.core.task import Task
from roc.idb.models.idb import IdbRelease
from roc.idb.models.idb import (
    ItemInfo,
    PacketHeader,
    PacketMapInfo,
    ParamInfo,
    TransferFunction,
)
from roc.idb.parsers import MIBParser
from roc.idb.parsers import PALISADEParser, SRDBParser
from roc.idb.parsers.idb_elements import Packet as IDBPacket
from roc.idb.parsers.idb_elements import ParameterDefinition, Enumeration
from roc.idb.tasks.load_palisade_metadata import put_palisade_metadata

__all__ = ["LoadTask"]


class IdbLoadTaskError(Exception):
    pass


class LoadTask(Task):
    plugin_name = "roc.idb"
    name = "idb_load_idb"

    @Connector.if_connected(settings.PIPELINE_DATABASE)
    def run(self):
        """
        Use the appropriate parser to load the selected IDB into the ROC database.
        """
        # get the database session
        self.session = self.pipeline.db.session

        # create the parser of IDB
        self.source = self.pipeline.args.idb_source
        if self.source == "SRDB":
            Parser = SRDBParser
        elif self.source == "PALISADE":
            Parser = PALISADEParser
        elif self.source == "MIB":
            Parser = MIBParser
        else:
            raise IdbLoadTaskError("Wrong parameter idb source")

        self.parser = Parser(self.pipeline.args.idb_path, self.pipeline.args.mapping)

        idb_version = self.parser.version()

        # check if the release already exists in the database
        try:
            _ = (
                self.session.query(IdbRelease)
                .filter_by(idb_version=idb_version, idb_source=self.source)
                .one()
            )

            logger.warning(
                f"IDB release ('{self.source}', '{idb_version}') already exists in the database"
            )
            return

        except NoResultFound:
            pass

        # parse the XML/dat files
        logger.debug(f"Start parsing using {self.source} parser")
        self.parser.parse()

        # put enum/tf into the database
        self._put_enumeration()
        self._put_tf()

        # put parameters into the database
        self._put_parameters()

        # put packets into the database
        self._put_packets()

        if self.source == "PALISADE":
            # load the PALISADE metadata
            put_palisade_metadata(self.session, self.parser)

        # update the database
        self.pipeline.db.update_database()

    def _get_idb(self):
        """
        To get the database_descr of the IDB to use.
        """
        # create the version of the IDB or retrieve it
        idb = get_or_create(
            self.session,
            IdbRelease,
            idb_version=self.parser.version(),
            idb_source=self.parser.source(),
            mib_version=self.parser.mib_version,
            current=False,
        )
        return idb

    def _put_enumeration(self):
        """
        Put Enumerations information into roc database
        """

        # get the idb
        idb = self._get_idb()

        enumerations_count = len(Enumeration.manager.instances)

        for i, enumeration in enumerate(Enumeration):
            print(
                "\rLoading enumerations {}%".format(
                    self._progress(i, enumerations_count)
                ),
                end="",
                flush=True,
            )

            srdb_id = enumeration.srdb_id
            palisade_id = enumeration.palisade_id

            # skip internal enumerations that uses PALISADE IDs instead of SRDB IDs
            if "_" in srdb_id:
                continue

            info = get_or_create(
                self.session,
                ItemInfo,
                srdb_id=srdb_id,
                palisade_id=palisade_id.strip()
                if isinstance(palisade_id, str)
                else palisade_id,
                idb=idb,
            )

            for eng, raw in enumeration.values.items():
                get_or_create(
                    self.session,
                    TransferFunction,
                    item_info_id=info.id_item_info,
                    raw=raw,
                    eng=eng,
                )
        logger.debug("END LOADING ENUMERATIONS IN DATABASE")

    def _put_tf(self):
        """
        Put transfer function information into roc database.
        """

        # get the idb
        idb = self._get_idb()

        tf_dict = self.parser.tf_dict

        tf_count = len(tf_dict)

        for i, srdb_id in enumerate(tf_dict):
            print(
                "\rLoading transfer functions {}%".format(self._progress(i, tf_count)),
                end="",
                flush=True,
            )

            # create the info

            info = get_or_create(
                self.session,
                ItemInfo,
                srdb_id=srdb_id,
                idb=idb,
            )

            for raw_eng_values in tf_dict[srdb_id]:
                get_or_create(
                    self.session,
                    TransferFunction,
                    item_info_id=info.id_item_info,
                    raw=raw_eng_values["raw"],
                    eng=raw_eng_values["eng"],
                )
        logger.debug("END LOADING TF IN DATABASE")

    def _put_packets(self):
        """
        Put packets SRDB information into roc database.
        """
        # get the idb
        idb = self._get_idb()

        # loop over registered packets
        packets = IDBPacket.manager.instances

        packets_count = len(packets)

        for i, packet in enumerate(packets):
            print(
                "\rLoading packets {}%".format(self._progress(i, packets_count)),
                end="",
                flush=True,
            )

            if packet.pid is None:
                print("\r", end="", flush=True)
                logger.warning("Packet {} does not have a PROCESS_ID".format(packet.id))

            palisade_id = packet.name
            try:
                srdb_id = self.parser.palisade_to_srdb_id[packet.type][palisade_id]
            except KeyError:
                srdb_id = None
                print("\r", end="", flush=True)
                logger.warning(
                    "Could not find {} in srdb/palisade mapping file".format(
                        palisade_id
                    )
                )

            # create the info
            info = get_or_create(
                self.session,
                ItemInfo,
                srdb_id=srdb_id,
                palisade_id=palisade_id.strip()
                if isinstance(palisade_id, str)
                else palisade_id,
                idb=idb,
            )

            # set the description of the packet
            info.item_descr = packet.description

            # if node is None, the packet is not defined in IDB
            if packet.node is None and self.parser is PALISADEParser:
                print("\r", end="", flush=True)
                logger.warning("{0} is not defined in IDB".format(packet.name))
                continue

            # create a representation of the packet
            pkt = get_or_create(
                self.session,
                PacketHeader,
                item_info_id=info.id_item_info,
                cat_eng=str(packet.category),
                cat=packet.category,
                packet_type=packet.type,
                byte_size=packet.byte_size,
                pid=packet.pid,
                pid_eng=str(packet.pid),
                spid=packet.spid,
                sid=packet.sid,
                service_type=packet.service_type,
                service_subtype=packet.service_subtype,
            )

            # now link parameters
            parameters = self._link_parameters(packet.parameters)

            # for each parameter and packet, create the map between them
            maps = []
            for parameter, pkt_param in zip(parameters, packet.parameters):
                # get the info if already present or create a new one
                info = get_or_create(
                    self.session,
                    PacketMapInfo,
                    parameter=parameter,
                    packet=pkt,
                )

                # update the information with what is in the database
                info.byte_position = pkt_param.byte_position
                info.bit_position = pkt_param.bit_position
                info.group_size = pkt_param.group_size
                info.block_size = pkt_param.block_size
                info.loop_value = pkt_param.group_counter

                # add to the maps
                maps.append(info)

            # add maps to the session for update
            self.session.add_all(maps)

    def _link_parameters(self, parameters):
        """
        Returns database representation of parameters given in argument.
        """
        # get the idb
        idb = self._get_idb()

        params = []
        for parameter in parameters:
            # create the info
            info = (
                self.session.query(ItemInfo)
                .filter_by(srdb_id=parameter.srdb_id, idb=idb)
                .one()
            )

            # create a representation of the parameter
            param = self.session.query(ParamInfo).filter_by(info=info).one()

            # add to list
            params.append(param)

        return params

    def _progress(self, index, total):
        return int((index + 1) / total * 100)

    def _put_parameters(self):
        """
        Put parameters SRDB information into roc database.
        """
        # get the idb
        idb = self._get_idb()

        # list of parameter definitions
        parameters = ParameterDefinition.manager.instances

        parameters_count = len(parameters)
        # loop over registered parameters
        for i, parameter in enumerate(parameters):
            print(
                "\rLoading parameters {}%".format(self._progress(i, parameters_count)),
                end="",
                flush=True,
            )

            palisade_id = parameter.palisade_id

            # create the info
            if parameter.srdb_id:
                packet_type, _, _ = self.parser.split_parameter_SRDBID(
                    parameter.srdb_id
                )

                assert packet_type in ["TM", "TC"]
                info = get_or_create(
                    self.session,
                    ItemInfo,
                    srdb_id=parameter.srdb_id,
                    palisade_id=palisade_id.strip()
                    if isinstance(palisade_id, str)
                    else palisade_id,
                    idb=idb,
                )

                # add the description of the parameter
                info.item_descr = parameter.description

                # get the associated transfer function if any
                if parameter.transfer_function:
                    # FIXME: handle multiple TF for one Palisade ID
                    if isinstance(parameter.transfer_function, list):
                        tf_srdb_id = parameter.transfer_function[0]
                    else:
                        tf_srdb_id = parameter.transfer_function
                    transfer_function = (
                        self.session.query(ItemInfo)
                        .filter_by(srdb_id=tf_srdb_id, idb=idb)
                        .filter(ItemInfo.srdb_id.like("%" + packet_type))
                        .first()
                    )
                    if transfer_function is None:
                        logger.warning(
                            f"No transfer function found for the {packet_type} {parameter.srdb_id} (Expected TF id: {parameter.transfer_function})"
                        )

                else:
                    transfer_function = None

                # create a representation of the parameter
                get_or_create(
                    self.session,
                    ParamInfo,
                    info=info,
                    par_bit_length=parameter.data_type.bit_length,
                    par_type=parameter.data_type.numpy_type,
                    par_max=numpy.float64(parameter.maximum),
                    par_min=numpy.float64(parameter.minimum),
                    par_unit=parameter.unit,
                    par_def=parameter.default,
                    par_is_editable=parameter.is_editable,
                    transfer_function=transfer_function,
                )
            else:
                logger.warning(f"No SRDB ID found for the parameter {palisade_id}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import json

import numpy
from poppy.core.logger import logger

from .parse_sequence_mixin import ParseSequenceMixin
from roc.idb.parsers.srdb_parser import SRDBParser

__all__ = ["MIBParser"]


class MIBParser(SRDBParser, ParseSequenceMixin):
    """
    Class used for the parsing of the .dat files of the IDB to get information on the
    IDB and MIB.
    """

    def setup_source_dir(self, dir_path):
        self._source = "MIB"

        # MIB files path
        self.dir_path = dir_path

    def __init__(self, dir_path, mapping_file_path):
        """
        Store the mib_parser path of the .dat files containing the IDB.
        """

        if mapping_file_path is not None:
            super().__init__(dir_path, mapping_file_path)
        else:
            self.setup_source_dir(dir_path)
            logger.warning("No mapping file supplied, IDB parsing is disabled")
            logger.info("Note: the mapping file is not needed to load sequences")

    def version(self):
        """
        Return the version of the IDB defined in the dat file.
        """

        vdf_filepath = os.path.join(self.dir_path, "vdf.dat")

        vdf_dtype = numpy.dtype(
            [
                ("NAME", "U8"),  # short version name (8 char)
                ("COMMENT", "U32"),
                ("DOMAINID", int),
                ("RELEASE", int),
                ("ISSUE", int),
            ]
        )
        vdf_table = numpy.loadtxt(vdf_filepath, delimiter="\t", dtype=vdf_dtype)

        # read the mib_version directly in the file
        with open(vdf_filepath) as f:
            self.mib_version = f.readline()

        try:
            # read the release note to compute the human readable version from the release date
            release_note_filepath = os.path.join(
                self.dir_path, "..", "release_notes.json"
            )

            with open(release_note_filepath) as json_file:
                release_note = json.load(json_file)
                version = release_note["mib_release_date"].replace("-", "")
        except Exception as e:
            # compute the default version
            logger.warning("Error during release note loading")
            logger.warning(e)
            version = f'{vdf_table["NAME"]}_{vdf_table["COMMENT"]}'

        return version

    def compute_global_param_group_size(self, global_parameter_srdb_id):
        """
        Compute the global and sub-parameters block contribution to the group size.

        # packet structure:
        # =======================
        # param 1
        # param 2
        # ------ global param 3 -
        # > sub-parm 3.1           |
        # > sub-pram 3.2           | sub_param_list_len
        # > sub-pram 3.3           |
        # -----------------------
        # param 4
        # param 5
        # =======================

        :param global_parameter_srdb_id:
        :return:
        """
        sub_param_list_len = len(self.global_parameters[global_parameter_srdb_id])

        return sub_param_list_len - 2

    def parse_and_load_sequences(self, session):
        mib_tables = self.parse_sequences()
        self.load_from_MIB(session, *mib_tables)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import argparse
from datetime import datetime

from poppy.core.command import Command
from poppy.core.logger import logger

from roc.idb.constants import (
    TIME_ARG_STRFORMAT,
    VALIDITY_MIN,
    VALIDITY_MAX,
    IDB_SOURCE,
    IDB_INSTALL_DIR,
    TIMEOUT,
    MIB_GITLAB_TOKEN,
    IDB_SVN_USER,
    IDB_SVN_PASSWORD,
)
from roc.idb.tasks import (
    LoadTask,
    LoadIdbDump,
    InstallIdbTask,
    SetCurrentTask,
    SetTrangeTask,
    ClearIdbReleaseTask,
    LoadPalisadeMetadataTask,
)

from .export import Export, ExportSequences
from .list import List, ListReleases
from .load import Load, LoadIdb, LoadSequences

__all__ = [
    "Idb",
    "Install",
    "SetCurrent",
    "Load",
    "LoadIdb",
    "LoadIdbDumpCommand",
    "LoadSequences",
    "List",
    "ListReleases",
    "SetTrangeTask",
    "Export",
    "ExportSequences",
    "LoadPalisadeMetadataTask",
]


class Idb(Command):
    """
    Parent command of the IDB commands.
    """

    __command__ = "idb"
    __command_name__ = "idb"
    __parent__ = "master"
    __parent_arguments__ = ["base"]
    __help__ = "Commands used to manage the IDB"

    def add_arguments(self, parser):
        """
        Add input arguments common to all the IDB plugin.

        :param parser: high-level pipeline parser
        :return:
        """

        parser.add_argument(
            "-t",
            "--access-token",
            type=str,
            default=[MIB_GITLAB_TOKEN],
            nargs=1,
            help="Gitlab access token",
        )
        parser.add_argument(
            "-T",
            "--timeout",
            help="Process timeout in seconds",
            type=int,
            nargs=1,
            default=[TIMEOUT],
        )

        parser.add_argument(
            "--force",
            action="store_true",
            default=False,
            help="Force IDB source files downloading (even if local files are already found)",
        )


class Install(Command):
    """
    Command to load the information of the IDB into the ROC database.
    """

    __command__ = "idb_install"
    __command_name__ = "install"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "Install the requested IDB in the given directory"

    def add_arguments(self, parser):
        parser.add_argument(
            "-s",
            "--idb-source",
            help="""
            IDB type: SRDB or PALISADE or MIB (default: %(default)s).
            """,
            choices=["MIB", "SRDB", "PALISADE"],
            default="MIB",
        )

        parser.add_argument(
            "-i",
            "--install-dir",
            help="""
            Path to the install directory (e.g. 'lib/idb/').
            """,
            type=str,
            default=None,
        )

        parser.add_argument(
            "-v",
            "--idb-version",
            help="""
            IDB version
            """,
            type=str,
            required=True,
        )

        parser.add_argument(
            "-l",
            "--load",
            help="load the IDB information into the ROC database",
            action="store_true",
        )

        parser.add_argument(
            "--validity-min",
            help=f"""
            Validity start time.
            Default value is {VALIDITY_MIN.strftime(TIME_ARG_STRFORMAT)}
            """,
            type=valid_time,
            default=VALIDITY_MIN.strftime(TIME_ARG_STRFORMAT),
        )

        parser.add_argument(
            "--validity-max",
            help=f"""
            Validity end time.
            Default value is {VALIDITY_MAX.strftime(TIME_ARG_STRFORMAT)}
            """,
            type=valid_time,
            default=VALIDITY_MAX.strftime(TIME_ARG_STRFORMAT),
        )

        parser.add_argument(
            "--current",
            action="store_true",
            default=False,
            help='Flag loaded IDB as "current" version to use',
        )

        parser.add_argument(
            "--svn-user",
            help="Username to login RPW IDB SVN repo",
            type=str,
            default=[IDB_SVN_USER],
            nargs=1,
        )

        parser.add_argument(
            "--svn-password",
            help="Password to login RPW IDB SVN repo",
            type=str,
            default=[IDB_SVN_PASSWORD],
            nargs=1,
        )

    def setup_tasks(self, pipeline):
        """
        Load the information from the IDB in the given XML files and set them
        into the ROC database.
        """
        from poppy.core.conf import settings

        # get pipeline args
        args = pipeline.args

        if not args.install_dir:
            # by default, install the idb in the lib folder of the pipeline
            args.install_dir = os.path.join(settings.ROOT_DIRECTORY, "lib", "idb")

        # instantiate the install task
        install_task = InstallIdbTask()

        if args.load:
            # update args
            if args.idb_source in ["PALISADE", "SRDB"]:
                # PALISADE case
                idb_dir_name = (
                    args.idb_version
                    if args.idb_version.startswith("V")
                    else "V" + args.idb_version
                )
                idb_path = os.path.join(args.install_dir, idb_dir_name)
                args.idb_path = os.path.join(idb_path, "IDB")
                args.mapping = os.path.join(
                    args.idb_path,
                    "xml",
                    "IDB_SRDB_Mapping_Table_by_MapGen.xml",
                )

            else:
                # MIB case
                idb_dir_name = args.idb_version
                idb_path = os.path.join(args.install_dir, idb_dir_name)
                args.mapping = os.path.join(idb_path, "idb_srdb_mapping_table.xml")
                args.idb_path = os.path.join(idb_path, "data")

            # prepare the tasks
            load_task = LoadTask()
            tasks = install_task | load_task

            if args.validity_min or args.validity_max:
                tasks = tasks | SetTrangeTask()

            if args.current:
                tasks = tasks | SetCurrentTask()

        else:
            tasks = install_task

        pipeline | tasks

        # set the pipeline for this situation

        pipeline.start = install_task


class LoadIdbDumpCommand(Command):
    """
    Command to load IDB SQL dump file into the ROC database.
    Required psql command line tool to work.
    """

    __command__ = "idb_load_idb_dump"
    __command_name__ = "load_idb_dump"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "Load IDB SQL dump file into the ROC database using psql tool."

    def add_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--dump-file",
            help="""
             Path of the IDB SQL dump file.
             Can be a local path or a URL.
             """,
            type=str,
            nargs=1,
            default=[None],
        )
        parser.add_argument(
            "-i",
            "--install-dir",
            help="""
            Path to the install directory.
            """,
            type=str,
            default=[IDB_INSTALL_DIR],
            nargs=1,
        )
        parser.add_argument(
            "-a",
            "--auth",
            type=str,
            default=(None, None),
            nargs=2,
            help="Authentication parameters (username, password)",
        )
        parser.add_argument(
            "--pg_restore",
            action="store_true",
            default=False,
            help="Use pg_restore tool instead of psql to load idb dump file."
            "(Required for PostgreSQL custom-format dump.)",
        )
        parser.add_argument(
            "-C",
            "--clear-temp",
            action="store_true",
            default=False,
            help="Delete temporary files, if any, at the end",
        )
        parser.add_argument(
            "--shell",
            action="store_true",
            default=False,
            help='Use "shell=True/False" option in subprocess.call()',
        )

    def setup_tasks(self, pipeline):
        pass

        load_idb_dump = LoadIdbDump()

        # Define pipeline workflow
        pipeline | load_idb_dump

        # set the pipeline first task
        pipeline.start = load_idb_dump


class ClearRelease(Command):
    """
    Command to clear a given IDB release from the database
    """

    __command__ = "clear_release"
    __command_name__ = "clear_release"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = (
        "Command to clear a given IDB release from the database "
        "(palisade_metadata table entries and IDB source files will not be deleted)"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "idb_version",
            help="""
            IDB version of the release to delete
            """,
            type=str,
        )

        parser.add_argument(
            "-s",
            "--idb-source",
            help="""
            IDB source of the release to delete
            """,
            type=str,
            default=None,
            required=True,
        )

        parser.add_argument(
            "--force",
            help="""
            Force deletion of the release (i.e., no interactive process)
            """,
            action="store_true",
            default=False,
        )

    def setup_tasks(self, pipeline):
        """
        Clear of a given IDB.
        """

        # the task
        clear_idb = ClearIdbReleaseTask()

        # build the pipeline workflow for the command
        pipeline | clear_idb
        pipeline.start = clear_idb


class SetCurrent(Command):
    """
    Command to set a given idb version as current
    """

    __command__ = "set_current"
    __command_name__ = "set_current"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "Command to set a given idb version as current"

    def add_arguments(self, parser):
        parser.add_argument(
            "idb_version",
            help="""
            IDB version of the release to set as current
            """,
            type=str,
        )

        parser.add_argument(
            "--idb-source",
            help="""
            IDB source (MIB, SRDB or PALISADE)
            """,
            type=str,
            default=None,
        )

    def setup_tasks(self, pipeline):
        """
        Load the information from the IDB in the given XML files and set them
        into the ROC database.
        """

        # the task
        set_current = SetCurrentTask()

        # set the pipeline for this situation
        pipeline | set_current
        pipeline.start = set_current


class SetTimeRange(Command):
    """
    Command to set the validity time range of a given idb
    """

    __command__ = "set_trange"
    __command_name__ = "set_trange"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "Command to set the validity time range of a given idb"

    def add_arguments(self, parser):
        parser.add_argument(
            "idb_version",
            help="""
            IDB version of the release to set
            """,
            type=str,
        )

        parser.add_argument(
            "-s",
            "--idb-source",
            help="""
            IDB source
            """,
            type=str,
            default="MIB",
        )

        parser.add_argument(
            "--validity-min",
            help="Validity start time. " 'Valid format is "YYYY-MM-DDTHH:MM:SS".',
            type=valid_time,
        )

        parser.add_argument(
            "--validity-max",
            help="Validity end time. " 'Valid format is "YYYY-MM-DDTHH:MM:SS".',
            type=valid_time,
        )

    def setup_tasks(self, pipeline):
        """
        Set validity time range of a given IDB.
        """

        # the task
        set_trange = SetTrangeTask()

        # set the pipeline for this situation
        pipeline | set_trange
        pipeline.start = set_trange


class GetLatestIdb(Command):
    """
    Command to get the latest release of a given IDB source
    """

    __command__ = "get_latest"
    __command_name__ = "get_latest"
    __parent__ = "idb"
    __parent_arguments__ = ["base"]
    __help__ = "Command to get the latest release of a given IDB source"

    def add_arguments(self, parser):
        parser.add_argument(
            "idb_source",
            help=f"""
            IDB source. Default is {IDB_SOURCE}.
            """,
            type=str,
            default=IDB_SOURCE,
        )

    def setup_tasks(self, pipeline):
        """
        Setup tasks workflow for command.
        """

        idb_source = pipeline.args.idb_source.upper()
        latest_idb = SetCurrentTask.get_latest_idb(idb_source)
        logger.info(
            f'Latest version "{latest_idb.idb_version}" found for IDB "{latest_idb.idb_source}"'
        )
        logger.info(
            f'Valid between "{latest_idb.validity_min}" and "{latest_idb.validity_max}"'
        )
        if idb_source == "MIB":
            logger.info(f'MIB Version is "{latest_idb.mib_version}"')

        # TODO - Add a way to avoid ERROR message
        #       "The pipeline is not linked to an entry point task, and cannot be run."


def valid_time(t, format=TIME_ARG_STRFORMAT):
    """
    Validate input datetime string format.

    :param t: input datetime string
    :param format: expected datetime string format
    :return: datetime object with input datetime info
    """
    if t:
        try:
            return datetime.strptime(t, format)
        except ValueError:
            msg = f"Not a valid date: '{t}'!"
            logger.error(msg)
            raise argparse.ArgumentTypeError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import shutil

import pytest

from poppy.core.test import CommandTestCase
from poppy.core.logger import logger

from roc.idb.models.idb import IdbRelease
from .idb_test_util import reset_db, load_idb


class TestSetTrangeCommand(CommandTestCase):
    """
    Test the set_trange command of roc.idb plugin.

    """

    # @pytest.mark.skip
    @pytest.mark.parametrize(
        "idb_source,idb_version,validity_min,validity_max",
        [
            ("MIB", "20200113", "2020-01-13T00:00:00", "2050-12-31T23:59:59"),
        ],
    )
    def test_set_trange(self, idb_source, idb_version, validity_min, validity_max):
        # Initializing database
        logger.debug("Reset database ...")
        reset_db(self)

        # Installing IDB
        logger.debug(f"Installing IDB [{idb_source}-{idb_version}] ...")
        self.install_dir = load_idb(self, idb_source, idb_version)

        # Command to test
        command_to_test = [
            "pop",
            "idb",
            "set_trange",
            idb_version,
            "-s",
            idb_source,
            "--validity-min",
            validity_min,
            "--validity-max",
            validity_max,
            "-ll",
            "INFO",
        ]

        # Run command to test
        self.run_command(command_to_test)

        # Verify expected behaviour
        expected_values = (
            self.session.query(IdbRelease.validity_min, IdbRelease.validity_max)
            .filter(
                IdbRelease.idb_version == idb_version,
                IdbRelease.idb_source == idb_source,
            )
            .one()
        )

        time_strformat = "%Y-%m-%dT%H:%M:%S"
        assert (expected_values[0].strftime(time_strformat) == validity_min) and (
            expected_values[1].strftime(time_strformat) == validity_max
        )

    def teardown_method(self, method):
        """
        Method called immediately after the test method has been called and the result recorded.

        This is called even if the test method raised an exception.

        :param method: the test method
        :return:
        """

        # rollback the database
        super().teardown_method(method)

        # Remove IDB folder
        shutil.rmtree(self.install_dir)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Reference documentation
# ROC-GEN-SYS-NTT-00038-LES_Iss01_Rev02(Mission_Database_Description_Document)

from sqlalchemy.ext.declarative import DeferredReflection

from .base import Base

__all__ = [
    "Sequence",
    "Procedure",
    "TelecommandParameter",
    "Telecommand",
    "StepSeparator",
    "RpwStateDescription",
]


class RpwStateDescription(DeferredReflection, Base):
    """
    RpwStateDescription model
    """

    __tablename__ = "figaro_rpwstatedescription"
    __table_args__ = {
        "schema": "music",
    }


class Procedure(DeferredReflection, Base):
    """
    Procedure model
    """

    __tablename__ = "figaro_procedure"
    __table_args__ = {
        "schema": "music",
    }


class Sequence(DeferredReflection, Base):
    """
    Sequence model, related to :model:`figaro.Procedure`
    """

    __tablename__ = "figaro_sequence"
    __table_args__ = {"schema": "music"}


class Telecommand(DeferredReflection, Base):
    """
    Telecommand statement model, related to :model:`figaro.Sequence`
    """

    __tablename__ = "figaro_telecommand"
    __table_args__ = {"schema": "music"}


class TelecommandParameter(DeferredReflection, Base):
    """
    Telecommand parameter model, related to :model:`figaro.Telecommand`
    """

    __tablename__ = "figaro_telecommandparameter"
    __table_args__ = {"schema": "music"}


class StepSeparator(DeferredReflection, Base):
    """
    Step separator statement model, related to :model:`figaro.Sequence`
    """

    __tablename__ = "figaro_stepseparator"
    __table_args__ = {"schema": "music"}


class TelecommandDuration(DeferredReflection, Base):
    """
    Telecommand duration model, related to :model:`figaro.TelecommandDuration`
    """

    __tablename__ = "figaro_telecommandduration"
    __table_args__ = {"schema": "music"}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.conf import settings
from poppy.core.db.connector import Connector
from poppy.core.db.handlers import create, get_or_create, get_update_or_create
from poppy.core.logger import logger
from poppy.core.target import PyObjectTarget
from poppy.core.task import Task
from roc.idb.parsers.mib_parser.procedure import parse_time_tag

__all__ = ["LoadSequencesTask"]

try:
    MUSIC_DATABASE = settings.MUSIC_DATABASE
except AttributeError:
    logger.warning("MUSIC_DATABASE is not defined in pipeline settings")
    MUSIC_DATABASE = "___MUSIC_DATABASE_NOT_DEFINED___"


def create_telecommand(
    session,
    *,
    position,
    sequence_id,
    srdb_id,
    comment,
    duration,
    parameters_data,
):
    """
    Create a telecommand and associated parameters in the database

    :param position:
    :param sequence_id:
    :param srdb_id:
    :param comment:
    :param duration:
    :param parameters_data:
    :return: the telecommand instance
    """

    from roc.idb.models.music.figaro import Telecommand, TelecommandParameter

    # create the telecommand
    telecommand = create(
        session,
        Telecommand,
        position=position,
        sequence_id=sequence_id,
        srdb_id=srdb_id,
        comment=comment,
        duration=duration,
        manual_dispatch=True,
    )

    for param_idx, parameter_data in enumerate(parameters_data):
        create(
            session,
            TelecommandParameter,
            position=param_idx,
            statement_id=telecommand.id_telecommand,
            is_formal_parameter=parameter_data["is_formal_parameter"],
            representation=parameter_data["representation"],
            srdb_id=parameter_data["srdb_id"],
            value=parameter_data["value"],
        )

    return telecommand


@PyObjectTarget.input("sequences")
@PyObjectTarget.input("tc_duration_table")
@PyObjectTarget.input("sequence_description_table")
@PyObjectTarget.input("rpw_states")
@PyObjectTarget.input("idb_version")
@PyObjectTarget.input("idb_source")
@Task.as_task(plugin_name="roc.idb", name="idb_load_sequences")
@Connector.if_connected(MUSIC_DATABASE)
def LoadSequencesTask(self):
    from roc.idb.models.music.figaro import (
        Procedure,
        Sequence,
        StepSeparator,
        Telecommand,
        TelecommandParameter,
        TelecommandDuration,
        RpwStateDescription,
    )
    from sqlalchemy.ext.declarative import DeferredReflection

    # get the meb connector and get the database
    connector = Connector.manager[settings.MUSIC_DATABASE]
    database = connector.get_database()

    # create music tables reflection and mappings
    DeferredReflection.prepare(database.engine)

    # ensure database is connected
    database.connectDatabase()

    # get a database session
    session = database.session_factory()

    sequences_data_dict = self.inputs["sequences"].value
    tc_duration_table = self.inputs["tc_duration_table"].value
    sequence_description_table = self.inputs["sequence_description_table"].value
    rpw_states = self.inputs["rpw_states"].value

    # load RPW modes and config
    get_update_or_create(
        session,
        RpwStateDescription,
        idb_version=self.inputs["idb_version"].value,
        idb_source="MIB",
        update_fields=rpw_states,
    )

    # load the tc durations in the database
    for srdb_id in tc_duration_table:
        get_update_or_create(
            session,
            TelecommandDuration,
            srdb_id=srdb_id,
            update_fields=tc_duration_table[srdb_id],
        )

    # load the sequences in the database
    for procedure_name in sequences_data_dict:
        # get or create the procedure
        procedure = get_or_create(
            session,
            Procedure,
            name=procedure_name,
        )

        for sequence_name, sequence_data in sequences_data_dict[procedure_name].items():
            # get and update the existing sequence
            # or create a new one

            if sequence_name in sequence_description_table:
                sequence_descriptions = sequence_description_table[sequence_name]
            else:
                sequence_descriptions = {
                    "long_description": sequence_data["description"],
                    "description": sequence_data["description"],
                }

            sequence = get_update_or_create(
                session,
                Sequence,
                name=sequence_name,
                update_fields={
                    "procedure_id": procedure.id_procedure,
                    "idb_source": "MIB",
                    "idb_version": self.inputs["idb_version"].value,
                    **sequence_descriptions,
                },
            )

            # delete existing statements with associated parameters
            tc_query = session.query(Telecommand).filter_by(
                sequence_id=sequence.id_sequence
            )
            for telecommand in tc_query:
                session.query(TelecommandParameter).filter_by(
                    statement_id=telecommand.id_telecommand
                ).delete()
                session.delete(telecommand)

            session.query(StepSeparator).filter_by(
                sequence_id=sequence.id_sequence
            ).delete()
            session.commit()

            # create new statements
            for statement_idx, statement_data in enumerate(
                sequence_data["statements"][:-1]
            ):
                # get the srdb_id
                srdb_id = statement_data["srdb_id"]

                # compute durations
                duration = parse_time_tag(
                    sequence_data["statements"][statement_idx + 1]["deltatime"]
                )
                min_duration = tc_duration_table.get(srdb_id, {"duration": 1})[
                    "duration"
                ]

                # compare the MIB duration with the min one (from the tc duration table)
                if duration < min_duration:
                    logger.warning(
                        f"Statement duration for statement '{sequence_name}:{srdb_id}' is lower than the "
                        f"value of the tc duration table ('{duration}' < '{min_duration}'). Maybe a "
                        f"spacecraft sequence ? Overriding MIB value"
                    )
                    statement_duration = min_duration
                else:
                    statement_duration = duration

                # create the telecommand
                create_telecommand(
                    session,
                    position=statement_idx * 2,
                    sequence_id=sequence.id_sequence,
                    srdb_id=statement_data["srdb_id"],
                    comment=statement_data["comment"],
                    duration=statement_duration,
                    parameters_data=statement_data["parameters"],
                )

                # create the step separator
                create(
                    session,
                    StepSeparator,
                    position=statement_idx * 2 + 1,
                    sequence_id=sequence.id_sequence,
                )

            # handle last the last TC
            statement_idx = len(sequence_data["statements"]) - 1
            statement_data = sequence_data["statements"][statement_idx]

            # get the last TC duration
            last_tc_duration = tc_duration_table.get(
                statement_data["srdb_id"], {"duration": 1}
            )["duration"]

            create_telecommand(
                session,
                position=statement_idx * 2,
                sequence_id=sequence.id_sequence,
                srdb_id=statement_data["srdb_id"],
                comment=statement_data["comment"],
                duration=last_tc_duration,
                parameters_data=statement_data["parameters"],
            )



# -*- coding: utf-8 -*-
import re

representation_map = {
    "E": "Eng",
    "H": "Hex",
    "D": "Def",
    "R": "Dec",
    "F": "FP",
    "S": "FP-like",
}


def parse_time_tag(time_tag):
    time_pattern = re.compile(r"([0-9]{2})\.([0-9]{2})\.([0-9]{2})")

    if time_tag == "":
        value = None
    elif time_pattern.match(str(time_tag)) is not None:
        hour, minute, second = time_tag.split(".")
        value = int(second) + int(minute) * 60 + int(hour) * 3600
    else:
        raise Exception("Invalid time tag field")

    return value


def seconds_to_hours_minutes_seconds(seconds):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return h, m, s


class Procedure(object):
    """Procedure class"""

    @staticmethod
    def create_from_MIB(csf_table, css_table, sdf_table, csp_table):
        """
        Create a procedure list from the MIB tables

        :param csf_table: the CSF table describes the top level of the sequence hierarchy (the sequence itself)
        :param css_table: the CSS table contains an entry for each element in the sequence
        :param sdf_table: the SDF table defines the values of the editable element parameters
        :param csp_table: the CSP defines the formal parameters associated with the current sequence
        :return: procedure list
        """

        # parse the formal parameters
        formal_parameters = {}

        for formal_parameter in csp_table:
            if formal_parameter["VTYPE"] == "R":
                if formal_parameter["RADIX"] == "H":
                    representation = "Hex"
                else:
                    representation = "Dec"
            else:
                representation = representation_map.get(formal_parameter["VTYPE"], None)

            formal_parameters[formal_parameter["FPNAME"]] = {
                "value": formal_parameter["DEFVAL"],
                "representation": representation,
                "fp_data": formal_parameter,
            }

        # parse the parameters and merge fp data if any
        parameters = {}

        for parameter in sdf_table:
            key = f'{parameter["SQNAME"]}-{parameter["ENTRY"]}-{parameter["ELEMID"]}'

            # if it's a FP
            if parameter["VTYPE"] == "F":
                fp_name = parameter["VALUE"]
                formal_parameter = formal_parameters[fp_name]

                value = formal_parameter["value"]
                representation = formal_parameter["representation"]

                if representation == "Hex" and value.startswith("0x"):
                    value = value[2:]

                parameters.setdefault(key, []).append(
                    {
                        "position": int(parameter["POS"]),
                        "representation": representation,
                        "value": value,
                        "is_formal_parameter": True,
                        "srdb_id": parameter["PNAME"],
                        "fp_name": fp_name,
                    }
                )
            # if the VTYPE is S
            # the value is stored in the csp table (like the fp)
            # we treat the parameter as if it were a formal parameter
            elif parameter["VTYPE"] == "S":
                # get the fp like identifier
                fp_like_name = parameter["PNAME"]

                # get the data
                formal_parameter_like_data = formal_parameters[fp_like_name]

                value = formal_parameter_like_data["value"]
                representation = formal_parameter_like_data["representation"]

                if representation == "Hex" and value.startswith("0x"):
                    value = value[2:]

                parameters.setdefault(key, []).append(
                    {
                        "position": int(parameter["POS"]),
                        "representation": representation,
                        "value": value,
                        "is_formal_parameter": False,
                        "srdb_id": parameter["PNAME"],
                    }
                )
            else:
                representation = representation_map.get(parameter["VTYPE"], None)

                value = parameter["VALUE"]

                if representation == "Hex" and value.startswith("0x"):
                    value = value[2:]

                parameters.setdefault(key, []).append(
                    {
                        "position": int(parameter["POS"]),
                        "value": value,
                        "representation": representation,
                        "is_formal_parameter": False,
                        "srdb_id": parameter["PNAME"],
                    }
                )

        # parse the statements and merge parameters data if any
        statements = {}

        for statement in css_table:
            sequence_name = statement["SQNAME"]
            key = f'{statement["SQNAME"]}-{statement["ENTRY"]}-{statement["ELEMID"]}'
            statements.setdefault(sequence_name, []).append(
                {
                    "srdb_id": statement["ELEMID"],
                    "deltatime": statement["RELTIME"],
                    "comment": statement["COMM"],
                    "position": int(statement["ENTRY"]),
                    # get parameters if any
                    "parameters": parameters.get(key, []),
                }
            )

        # create the dictionary of procedures and sequences
        procedure_data = {}

        # populate the dict with sequences data
        for sequence in csf_table:
            procedure = procedure_data.setdefault(sequence["DOCNAME"], {})
            sequence_name = sequence["SQNAME"]

            procedure[sequence_name] = {
                "statements": statements[sequence_name],
                "description": sequence["DESC"],
            }

        return procedure_data

    @staticmethod
    def as_xls_data(sequence_data, tc_duration_table):
        stmt_header = [
            "Stmt_nr",
            "Step_nr",
            "Stmt_type",
            "Stmt_id",
            "Blk_flag",
            "Time_tag",
            "info",
            "Param_val_int_tm",
            "Param_val_tm",
            "Proforma",
            "Packet",
            "Manual_Dispatch",
        ]

        cmd_params_header = [
            "Cmd_nr",
            "Param_name",
            "Param_val_int",
            "Param_val",
            "FP_def_val_int",
            "FP_def_val",
        ]

        tlm_values_header = [
            "TLM_nr",
            "val_relation",
            "param_val_int",
            "param_val",
        ]

        pkt_params_header = [
            "Pkt_nr",
            "Param_name",
            "Param_val_int",
            "Param_val",
        ]

        info_header = [
            "Procedure Title",
            "Objective",
            "Summary of Constraints",
            "Mode at start",
            "Intermediate Mode (if any)",
            "Mode at end",
            "Procedure Duration",
            "Short Sequence Description",
            "Long Sequence Description",
        ]

        stmt = {"header": stmt_header, "rows": []}
        cmd = {"header": cmd_params_header, "rows": []}

        # initialize the sequence duration
        duration = 0
        for statement in sequence_data["statements"]:
            statement_data = {
                "Stmt_nr": statement["position"],
                "Step_nr": statement["position"],
                "Stmt_type": "CMD",
                "Stmt_id": statement["srdb_id"],
                "Blk_flag": "",
                "Time_tag": statement["deltatime"].replace(":", "."),
                "info": statement["comment"],
                "Param_val_int_tm": "",
                "Param_val_tm": "",
                "Proforma": "",
                "Packet": "",
                "Manual_Dispatch": "Y",
            }

            stmt["rows"].append(statement_data)

            # increment the total duration
            duration += parse_time_tag(statement["deltatime"])

            for parameter in statement["parameters"]:
                if parameter["is_formal_parameter"]:
                    parameter_data = {
                        "Cmd_nr": statement["position"],
                        "Param_name": parameter["srdb_id"],
                        "Param_val_int": "FP",
                        "Param_val": parameter["fp_name"],
                        "FP_def_val_int": parameter["representation"],
                        "FP_def_val": parameter["value"],
                    }
                else:
                    parameter_data = {
                        "Cmd_nr": statement["position"],
                        "Param_name": parameter["srdb_id"],
                        "Param_val_int": parameter["representation"],
                        "Param_val": parameter["value"],
                    }
                cmd["rows"].append(parameter_data)

        # get the last TC duration
        last_tc_duration = tc_duration_table.get(
            stmt["rows"][-1]["Stmt_id"], {"duration": 1}
        )["duration"]

        # add the last TC duration
        duration += last_tc_duration

        description = sequence_data["description"]

        h, m, s = seconds_to_hours_minutes_seconds(duration)

        info = {
            "header": info_header,
            "rows": [
                {
                    "Procedure Title": description,
                    "Objective": description,
                    "Summary of Constraints": description,
                    "Mode at start": "",
                    "Intermediate Mode (if any)": "",
                    "Mode at end": "",
                    "Procedure Duration": f"{h}:{m:02d}:{s:02d}",
                    "Short Sequence Description": description,
                    "Long Sequence Description": description,
                }
            ],
        }

        # there are no tml nor pkt data in the MIB
        tlm = {"header": tlm_values_header, "rows": []}
        pkt = {"header": pkt_params_header, "rows": []}

        # return the data by sheet
        return {
            "STMT": stmt,
            "CMD Params": cmd,
            "TLM Values": tlm,
            "PKT Params": pkt,
            "Info": info,
        }

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return self.name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from poppy.core.configuration import Configuration
from poppy.core.db.dry_runner import DryRunner
from poppy.core.generic.cache import CachedProperty
from poppy.pop.db_connector import POPPy as POPPyConnector
from roc.idb.models.idb import IdbRelease
from roc.idb.manager import IDBManager
from sqlalchemy.orm.exc import NoResultFound

__all__ = ["IdbConnector"]


class IdbDatabaseError(Exception):
    """
    Errors for the connector of the POPPy database.
    """


class IdbConnector(POPPyConnector):
    """
    A class for querying the POPPy database with the IDB schema.
    """

    @CachedProperty
    def idb_manager(self):
        """
        Create the IDB manager and store it.
        """
        return IDBManager(self.session)

    @DryRunner.dry_run
    def get_idb(self):
        """
        Return the database_descr object of the selected version of the IDB.
        """
        if self._idb is None:
            raise ValueError("Must specify a version for the IDB")

        # construct the query
        query = self.session.query(IdbRelease)
        query = query.filter_by(idb_version=self._idb)
        try:
            return query.one()
        except NoResultFound:
            raise IdbDatabaseError(
                (
                    "The IDB version {0} was not found on the database. "
                    + "Have you loaded the IDB inside the database with "
                    + "help of the -> pop rpl load command?"
                ).format(self._idb)
            )

    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    def configuration(self, configuration):
        self._configuration = configuration

        # get the descriptor file
        descriptor = Configuration.manager["descriptor"]
        self._pipeline = descriptor["pipeline.release.version"]
        self._pipeline_name = descriptor["pipeline.identifier"]

        if "idb_version" in self._configuration:
            self._idb = self._configuration.idb_version



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import csv
import os
import re

import numpy

from poppy.core.logger import logger
from roc.idb.parsers.idb_elements import (
    Enumeration,
    Packet,
    Parameter,
    ParameterDefinition,
    drop_numpy_type,
)
from roc.idb.parsers.idb_parser import IDBParser

from . import dtypes

__all__ = ["SRDBParser"]


def vectorial_match(data, regex=""):
    r = re.compile(regex)

    @numpy.vectorize
    def vmatch(data):
        return bool(r.match(str(data)))

    return vmatch(data)


class SRDBParser(IDBParser):
    """
    Class used for the parsing of the .dat files of the IDB to get information on the
    IDB and SRDB.
    """

    @staticmethod
    def filter_table(table, regex_dict={r"SQNAME": r"AIWF\d{3}[A-Z]"}):
        """
        Filter tables to keep only specified values
        """
        match = numpy.logical_and.reduce(
            [
                vectorial_match(table[key], regex=value)
                for key, value in regex_dict.items()
            ]
        )

        return table[match]

    @classmethod
    def from_table(cls, filepath, dtype, regex_dict={}):
        """
        Load IDB info from the given csv file.

        :param filepath: Path to the csv file
        :param dtype: Numpy dtype of the table
        :param regex_dict: Dictionary use to filter the table, example: regex_dict={r"SQNAME": r"AIWF\d{3}[A-Z]"}
        :return:
        """

        # create a generic dtype to avoid parsing errors with empty values
        obj_dtype = numpy.dtype([(key, object) for key, _ in dtype.fields.items()])

        with open(filepath, newline="") as csv_file:
            csv_table_reader = csv.reader(csv_file, delimiter="\t")
            table = numpy.array(
                [tuple(row)[: len(dtype)] for row in csv_table_reader],
                dtype=obj_dtype,
            )
            if regex_dict:
                table = cls.filter_table(table, regex_dict=regex_dict)

            # force the dtype of the returned table
            return table.astype(dtype)

    def setup_source_dir(self, dir_path):
        self._source = "SRDB"

        # IDB files path
        self.dir_path = os.path.join(dir_path, self._source, "Dat")

    def __init__(self, dir_path, mapping_file_path):
        """
        Store the dir path of the .dat files containing the IDB.
        """

        self.setup_source_dir(dir_path)

        super().__init__(mapping_file_path)

        # Packet indexing using TPSD id (variable one)
        self.tpsd_mapping = dict()
        # packet indexing using spid (static one)
        self.spid_mapping = dict()

        # list of detailed parameter for each global one
        self.global_parameters = dict()

        # min and max value indexed by srdbid
        self.min_max_dict = dict()

    def version(self):
        """
        Return the version of the IDB defined in the dat file.
        """

        vdf_table = self.from_table(
            os.path.join(self.dir_path, "vdf.dat"), dtype=dtypes.vdf_dtype
        )
        # read the version information from the srdb (the version is included
        # in a long comment)
        vdf_version = vdf_table["COMMENT"].item().split(" ")
        # use the word "Version" as a tag to find the usefull information
        version_index = vdf_version.index("Version") + 1
        # store the version and remove the trailing quote
        srdb_version = vdf_version[version_index].strip("'")
        return srdb_version

    def enumeration_srdb(self, enumeration, kind):
        """
        Return the name of the enumeration in the SRDB according to the type
        given in argument.

        For the moment, for the SRDB parser, the enumeration name is the
        SRDBID whereas for PALISADE parser, it's the PALISADE ID
        """
        return enumeration.name

    def get_palisade_id(self, srdb_id):
        """
        Return the corresponding palisade ID.
        """
        return self.srdb_to_palisade_id[srdb_id]

    def add_packet(
        self,
        name,
        srdbid,
        kind=None,
        packet_type=None,
        pid=None,
        spid=None,  # SCOS Packet ID
        sid=None,
        packet_category=None,
        description=None,
        bit_size=0,
        service_type=None,
        service_subtype=None,
    ):
        packet = Packet(name, srdbid, kind=kind, type=packet_type, bit_size=bit_size)

        packet.pid = pid
        packet.pid_eng = str(pid)
        packet.spid = spid
        packet.sid = sid
        packet.sid_eng = str(sid)
        packet.service_type = service_type
        packet.service_type_eng = str(service_type)
        packet.service_subtype = service_subtype
        packet.service_subtype_eng = str(service_subtype)

        # get the packet category
        packet.category = packet_category
        packet.category_eng = str(packet_category)

        # get the description of the packet
        packet.description = description

        # create empty list of parameters
        packet.parameters = []

        return packet

    def create_mapping(self):
        # xml mapping
        super().create_mapping()

        # mapping between srdbid, short palisade ID and param/packet description

        sw_para_file_path = os.path.join(self.dir_path, "sw_para.dat")

        if not os.path.isfile(sw_para_file_path):
            sw_para_file_path = os.path.join(self.dir_path, "ASW_sw_para.dat")

        sw_para = self.from_table(sw_para_file_path, dtype=dtypes.sw_para_dtype)

        self.sw_para_mapping = {
            row["SCOS_NAME"]: (row["SW_NAME"], row["SW_DESCR"]) for row in sw_para
        }

    def add_palisade_prefix(self, short_palisade_id, srdb_id):
        """
        Save the correspondence between the palisade and the (long) srdb ID
        """
        if srdb_id in self.srdb_to_palisade_id:
            logger.debug(
                f"Comparing reconstructed palisade IDs ({self.srdb_to_palisade_id[srdb_id]}, "
                f"{short_palisade_id}) associated with this srdb ID ({srdb_id})"
            )
            if (
                self.srdb_to_palisade_id[srdb_id][3:].strip()
                != short_palisade_id.strip()
            ):
                logger.error(
                    f"The short palisade ID is not compatible with the long one "
                    f"({self.srdb_to_palisade_id[srdb_id]} != {short_palisade_id}) associated with "
                    f"this srdb ID ({srdb_id})"
                )

            return self.srdb_to_palisade_id[srdb_id]

    def create_parameter(
        self,
        palisade_id,
        srdb_id,
        ptc,
        pfc,
        parameter_type,
        kind,
        unit=None,
        enumeration=None,
        min_max_values=(None, None),
        transfer_function=None,
    ):
        _, description = self.sw_para_mapping[srdb_id]

        definition_dict = {
            "description": description,
            "palisade_id": palisade_id,
            "default": None,
            "is_editable": None,
            "transfer_function": transfer_function,
            "minimum": None,
            "maximum": None,
            "unit": unit,
        }

        if enumeration is not None:
            enumeration.data_type = repr(self.srdb_to_palisade_type(ptc, pfc))
            definition_dict["data_type"] = enumeration
        else:
            definition_dict["data_type"] = self.srdb_to_palisade_type(ptc, pfc)
            (
                definition_dict["minimum"],
                definition_dict["maximum"],
            ) = min_max_values

        definition = ParameterDefinition(srdb_id, **definition_dict)

        parameter = Parameter(definition, parameter_type)

        return parameter

    def convert_to_palisade(self, srdb_id):
        """Convert SRDB ID into PALISADE ID using mapping table"""
        if srdb_id in self.srdb_to_palisade_id:
            return self.srdb_to_palisade_id[srdb_id]
        else:
            logger.error("No PALISADE ID associated to SRDB ID : %s", srdb_id)

    def load_min_max_values(self):
        """
        Load min and max parameters value from ccs SRDB data
        """

        ccs_cap_dtype = numpy.dtype(
            [("NUMBR", "U10"), ("XVALS", "U17"), ("YVALS", "U17")]  # srdbid
        )

        prv_dtype = numpy.dtype(
            [("NUMBR", "U10"), ("MINVAL", "U64"), ("MAXVAL", "U64")]
        )

        prf_dtype = numpy.dtype(
            [
                ("NUMBR", "U10"),
                ("DESCR", "U24"),
                ("INTER", "U1"),
                ("DSPFMT", "U1"),
                ("RADIX", "U1"),
            ]
        )

        ccs_table = self.from_table(
            os.path.join(self.dir_path, "ccs.dat"),
            dtype=ccs_cap_dtype,
            regex_dict={"NUMBR": "^CIW.*"},
        )

        cap_table = self.from_table(
            os.path.join(self.dir_path, "cap.dat"),
            dtype=ccs_cap_dtype,
            regex_dict={"NUMBR": "^CIW.*"},
        )

        prv_table = self.from_table(
            os.path.join(self.dir_path, "prv.dat"),
            dtype=prv_dtype,
            regex_dict={"NUMBR": "^VIW.*"},
        )

        prf_table = self.from_table(
            os.path.join(self.dir_path, "prf.dat"),
            dtype=prf_dtype,
            regex_dict={"NUMBR": "^VIW.*"},
        )

        for row in ccs_table:
            if self.min_max_dict.get(row["NUMBR"]) is None:
                self.min_max_dict[row["NUMBR"]] = (row["XVALS"], row["XVALS"])
            else:
                min_value, max_value = self.min_max_dict[row["NUMBR"]]
                self.min_max_dict[row["NUMBR"]] = (
                    min(min_value, row["XVALS"], key=float),
                    max(max_value, row["XVALS"], key=float),
                )

        for row in cap_table:
            if self.min_max_dict.get(row["NUMBR"]) is None:
                self.min_max_dict[row["NUMBR"]] = (row["XVALS"], row["XVALS"])
            else:
                min_value, max_value = self.min_max_dict[row["NUMBR"]]
                self.min_max_dict[row["NUMBR"]] = (
                    min(min_value, row["XVALS"], key=float),
                    max(max_value, row["XVALS"], key=float),
                )

        # get the default min/max representation from the prf table
        #
        # 'D' – Decimal
        # 'H' – Hexadecimal
        # 'O' – Octal

        min_max_repr = {}

        for row in prf_table:
            min_max_repr[row["NUMBR"]] = row["RADIX"]

        for row in prv_table:
            # convert each min/max value into Dec
            if min_max_repr[row["NUMBR"]] == "H":
                min_value = str(int(row["MINVAL"], 16))
                max_value = str(int(row["MAXVAL"], 16))
            elif min_max_repr[row["NUMBR"]] == "O":
                min_value = str(int(row["MINVAL"], 8))
                max_value = str(int(row["MAXVAL"], 8))
            else:
                min_value = row["MINVAL"]
                max_value = row["MAXVAL"]

            self.min_max_dict[row["NUMBR"]] = (min_value, max_value)

    def add_tc_packet(self):
        """
        TC packet loading from ccf SRDB data
        """

        ccf_table = self.from_table(
            os.path.join(self.dir_path, "ccf.dat"),
            dtype=dtypes.ccf_dtype,
            regex_dict={"CNAME": "^ZIW.*"},
        )

        for row in ccf_table:
            # create the object of the packet
            short_name = row["DESCR"]
            logger.debug("Reading TC packet: %s" % short_name)

            packet_type, kind, srdbid = self.split_packet_SRDBID(row["CNAME"])
            process_id, packet_category = self.split_apid(row["APID"])

            name = self.add_palisade_prefix(short_name, row["CNAME"])
            service_type = row["TYPE"]
            service_subtype = row["STYPE"]

            self.add_packet(
                name,
                srdbid,
                kind=kind,
                packet_type=packet_type,
                pid=process_id,
                spid=None,
                sid=None,
                packet_category=packet_category,
                description=row["DESCR2"],
                service_type=service_type,
                service_subtype=service_subtype,
            )

    def add_tm_packet(self):
        """
        TM packet loading from pid SRDB data
        """

        pid_table = self.from_table(
            os.path.join(self.dir_path, "pid.dat"), dtype=dtypes.pid_dtype
        )

        tpcf_table = self.from_table(
            os.path.join(self.dir_path, "tpcf.dat"),
            dtype=numpy.dtype([("SPID", object), ("NAME", "U12"), ("SIZE", object)]),
            regex_dict={"NAME": "^YIW.*"},
        )

        spid_srdb_id_mapping = {}

        for tpcf_row in tpcf_table:
            spid_srdb_id_mapping[tpcf_row["SPID"]] = (
                tpcf_row["NAME"],
                tpcf_row["SIZE"],
            )

        for pid_row in pid_table:
            if pid_row["SPID"] not in spid_srdb_id_mapping:
                # skip non RPW packets
                continue

            srdb_id, size = spid_srdb_id_mapping[pid_row["SPID"]]

            # create the object of the packet
            short_name, description = self.sw_para_mapping[srdb_id]
            logger.debug("Reading TM packet: %s" % short_name)
            packet_type, kind, srdbid = self.split_packet_SRDBID(srdb_id)
            process_id, packet_category = self.split_apid(pid_row["APID"])

            name = self.add_palisade_prefix(short_name, srdb_id)
            service_type = pid_row["TYPE"]
            service_subtype = pid_row["STYPE"]

            # save packet size if it's not a variable structure (else set to 0
            # and fill it later)
            header_byte_size = 16
            byte_size = int(size) - header_byte_size if size != "" else 0
            packet = self.add_packet(
                name,
                srdbid,
                kind=kind,
                packet_type=packet_type,
                pid=process_id,
                spid=pid_row["SPID"],
                sid=pid_row["PI1_VAL"],
                packet_category=packet_category,
                description=description,
                bit_size=byte_size * 8,
                service_type=service_type,
                service_subtype=service_subtype,
            )
            # if Telemetry Packet Structure Definition (TPSD) exist, use it to
            # index packets
            if pid_row["TPSD"] > -1:
                self.tpsd_mapping[pid_row["TPSD"]] = packet
            # else it's a static and use spid to index the packet
            else:
                self.spid_mapping[pid_row["SPID"]] = packet

    def add_tc_parameter(self):
        """
        TC parameter loading from cpc SRDB data
        """

        cpc_table = self.from_table(
            os.path.join(self.dir_path, "cpc.dat"),
            dtype=dtypes.cpc_dtype,
            regex_dict={"PNAME": "^PIW.*"},
        )
        for row in cpc_table:
            srdb_id = row["PNAME"]
            parameter_type, kind, short_srdb_id = self.split_parameter_SRDBID(srdb_id)
            short_name = row["DESCR"]
            logger.debug("Reading TC parameter: %s" % short_name)

            palisade_id = self.add_palisade_prefix(short_name, row["PNAME"])
            enumeration = None
            min_max_values = (None, None)
            tf_srdb_id = None

            if row["PRFREF"] != "":
                min_max_values = self.min_max_dict[row["PRFREF"]]
            elif row["PAFREF"] != "":
                # if not a transfer function
                if row["PAFREF"][3] == "T":
                    enumeration = Enumeration.manager[row["PAFREF"]]
                else:
                    min_max_values = self.min_max_dict[row["PAFREF"]]
                tf_srdb_id = row["PAFREF"]
            elif row["CCAREF"] != "":
                # if not a transfer function
                if row["CCAREF"][3] == "T":
                    enumeration = Enumeration.manager[row["CCAREF"]]
                else:
                    min_max_values = self.min_max_dict[row["CCAREF"]]
                tf_srdb_id = row["CCAREF"]

            self.create_parameter(
                palisade_id,
                srdb_id,
                row["PTC"],
                row["PFC"],
                parameter_type,
                kind,
                unit=row["UNIT"],
                enumeration=enumeration,
                min_max_values=min_max_values,
                transfer_function=tf_srdb_id,
            )

    def add_tm_parameter(self):
        """
        TM parameter loading from pcf SRDB data
        """

        pcf_table = self.from_table(
            os.path.join(self.dir_path, "pcf.dat"),
            dtype=dtypes.pcf_dtype,
            regex_dict={"NAME": "^NIW.*"},
        )

        for row in pcf_table:
            srdb_id = row["NAME"]
            parameter_type, kind, short_srdb_id = self.split_parameter_SRDBID(srdb_id)
            short_name = row["DESCR"]
            logger.debug("Reading TM paramater: %s" % short_name)

            palisade_id = self.add_palisade_prefix(short_name, srdb_id)
            enumeration = None
            tf_srdb_id = None
            min_max_values = (None, None)
            if row["CURTX"] != "":
                # if not a transfer function
                if row["CURTX"][3] == "T":
                    enumeration = Enumeration.manager[row["CURTX"]]
                else:
                    min_max_values = self.min_max_dict[row["CURTX"]]
                tf_srdb_id = row["CURTX"]

            self.create_parameter(
                palisade_id,
                srdb_id,
                row["PTC"],
                row["PFC"],
                parameter_type,
                kind,
                unit=row["UNIT"],
                enumeration=enumeration,
                min_max_values=min_max_values,
                transfer_function=tf_srdb_id,
            )

    def add_tc_packetsubelement(self):
        """
        TC packet structure loading from cdf SRDB data
        """

        cdf_table = self.from_table(
            os.path.join(self.dir_path, "cdf.dat"),
            dtype=dtypes.cdf_dtype,
            regex_dict={"CNAME": "^ZIW.*"},
        )

        # size of and position in the group of parameters
        group_size = 0
        group_position = 0
        spare_size = 0

        for row in cdf_table:
            srdb_long_id = row["CNAME"]
            packet_type, kind, srdbid = self.split_packet_SRDBID(srdb_long_id)
            name = self.get_palisade_id(srdb_long_id)
            packet = Packet.manager[name]

            # if no parameter name check if it's a spare
            if row["PNAME"] == "":
                if row["DESCR"] == "SPARE":
                    packet.add_to_size(row["ELLEN"])
                    if group_size != 0:
                        # add the bit size of the spare to the total
                        spare_size += row["ELLEN"]
                        # decrement the group size counter since spares 0x0 do
                        # not count in palisade
                        group_size -= 1
                    continue

                else:
                    logger.error("Unknown object uncounter during SRDB parsing")
                    raise TypeError(
                        "palisade_id: %s, srdb_id: %s" % (name, srdb_long_id)
                    )
            else:
                parameter_srdb_id = row["PNAME"]
                parameter_palisade_id = self.get_palisade_id(parameter_srdb_id)
                definition = ParameterDefinition.manager[parameter_srdb_id]
                (
                    parameter_type,
                    parameter_kind,
                    _,
                ) = self.split_parameter_SRDBID(parameter_srdb_id)

                # update the default value of the parameter
                definition.default = row["VALUE"] if row["VALUE"] != "" else None

                # update the editable status and the max/min values

                if row["ELTYPE"] == "F":
                    # if the parameter is fixed, max = min = default
                    definition.maximum = definition.default
                    definition.minimum = definition.default

                definition.is_editable = row["ELTYPE"] == "E"

                parameter = Parameter(definition, parameter_type)
                # store the byte position and bit position inside the byte
                parameter.byte_position = row["BIT"] // 8
                parameter.bit_position = row["BIT"] % 8
                if parameter_kind != "global":
                    packet.add_to_size(parameter.definition.data_type.bit_length)
                    packet.parameters.append(parameter)

            if group_size != 0:
                if parameter_kind != "global":
                    if (group_size - 1) != group_position:
                        group_position += 1
                    else:
                        packet.set_group_properties(group_size, spare_size=spare_size)
                        # reset size and position
                        group_position = 0
                        group_size = row["GRPSIZE"]
                        spare_size = 0
                else:
                    # update the group size to take into account subparams
                    # and remove global parameter
                    len_detailed = len(self.global_parameters[parameter_srdb_id])
                    group_size += len_detailed - 1
            else:
                group_size = row["GRPSIZE"]
                spare_size = 0

            logger.debug(
                "Update TC packet structure %s with %s parameter"
                % (name, parameter_palisade_id)
            )

    def add_tm_packetsubelement(self):
        """
        TM packet structure loading from vpd SRDB data
        """

        vpd_table = self.from_table(
            os.path.join(self.dir_path, "vpd.dat"),
            dtype=dtypes.vpd_dtype,
            regex_dict={"NAME": "^NIW.*"},
        )

        # the bit position of the parameter
        start_cursor = 0
        bit_cursor = 0

        # size of and position in the group of parameters
        group_size = 0
        group_position = 0

        # reorder the table by TPSD and POS
        for row in numpy.sort(vpd_table, order=["TPSD", "POS"]):
            # retrieve the variable packet using the tpsd_mapping
            packet = self.tpsd_mapping[row["TPSD"]]

            # For the first parameter, reset the bit cursor
            if row["POS"] == 1:
                bit_cursor = 0
            bit_cursor += row["OFFSET"]
            long_srdb_id = repr(packet)

            name = self.get_palisade_id(long_srdb_id)
            # verify that the packet is already defined in the packet manager
            assert packet == Packet.manager[name]

            parameter_srdb_id = row["NAME"]

            definition = ParameterDefinition.manager[parameter_srdb_id]
            (parameter_type, parameter_kind, _) = self.split_parameter_SRDBID(
                parameter_srdb_id
            )

            parameter = Parameter(definition, parameter_type)

            # parameter.block_size =  bsize // 8  # in bytes

            # store the byte position and the bit position inside the byte
            parameter.byte_position = bit_cursor // 8
            parameter.bit_position = bit_cursor % 8

            # increment the byte position with the size of the parameter
            # and store it
            bit_cursor += definition.data_type.bit_length

            if parameter_kind != "global":
                # global parameters are not stored in the packet
                packet.update_size(bit_cursor)
                packet.parameters.append(parameter)
            # the group size (number of params in a group) is specified in the
            # parameter just before the beginning of the group
            if group_size != 0:
                if parameter_kind != "global":
                    if (group_size - 1) != group_position:
                        # move ahead in the group
                        group_position += 1
                    else:
                        packet.set_group_properties(
                            group_size, bit_cursor - start_cursor
                        )
                        # reset size and position
                        group_position = 0
                        group_size = row["GRPSIZE"]
                        start_cursor = bit_cursor
                else:
                    # update the group size to take into account the global parameter and the sub-params block
                    group_size += self.compute_global_param_group_size(
                        parameter_srdb_id
                    )

            else:
                group_size = row["GRPSIZE"]
                start_cursor = bit_cursor

            logger.debug(
                "Update TM packet structure %s with %s parameter (%s)"
                % (name, parameter_srdb_id, start_cursor)
            )

    def compute_global_param_group_size(self, global_parameter_srdb_id):
        """
        Compute the global and sub-parameters block contribution to the group size.

        # packet structure:
        # =======================
        # param 1
        # param 2
        # ------ global param 3 -
        # > sub-parm 3.1           |
        # > sub-pram 3.2           | sub_param_list_len
        # > sub-pram 3.3           |
        # -----------------------
        # param 4
        # param 5
        # =======================

        :param global_parameter_srdb_id:
        :return:
        """
        sub_param_list_len = len(self.global_parameters[global_parameter_srdb_id])
        return sub_param_list_len - 1

    def add_static_packetsubelement(self):
        """
        TM/TC static structure loading from plf SRDB data
        """

        plf_table = self.from_table(
            os.path.join(self.dir_path, "plf.dat"),
            dtype=dtypes.plf_dtype,
            regex_dict={"NAME": "^NIW.*"},
        )

        # reorder the table and loop over rows
        for row in numpy.sort(plf_table, order=["SPID", "OFFBY", "OFFBI"]):
            # since it's a static, use spid to retrieve the packets
            packet = self.spid_mapping[row["SPID"]]
            srdb_long_id = repr(packet)
            packet_type, kind, srdbid = self.split_packet_SRDBID(srdb_long_id)
            name = self.get_palisade_id(srdb_long_id)
            # verify that the packet is already defined in the packet manager
            assert packet == Packet.manager[name]

            parameter_srdb_id = row["NAME"]
            definition = ParameterDefinition.manager[parameter_srdb_id]
            (parameter_type, parameter_kind, _) = self.split_parameter_SRDBID(
                parameter_srdb_id
            )
            parameter = Parameter(definition, parameter_type)

            # store the byte position and bit position inside the byte
            header_byte_size = 16
            parameter.byte_position = row["OFFBY"] - header_byte_size
            parameter.bit_position = row["OFFBI"]

            if parameter_kind != "global":
                packet.parameters.append(parameter)
            logger.debug(
                "Update TM/TC static structure %s with %s parameter"
                % (name, parameter_srdb_id)
            )

    def create_parameters(self):
        # TC parameter in cpc SRDB data
        self.add_tc_parameter()
        logger.debug("TC parameter entries loaded.")

        # TM parameter in pcf SRDB data
        self.add_tm_parameter()
        logger.debug("TM parameter entries loaded.")

    def create_packets(self):
        # TC packet in ccf SRDB data
        self.add_tc_packet()
        logger.debug("TC packet entries loaded.")

        # TM packet in pid SRDB data
        self.add_tm_packet()
        logger.debug("TM packet entries loaded.")

    def populate_packets_with_parameters(self):
        # TC packet structure in cdf SRDB data
        self.add_tc_packetsubelement()
        logger.debug("TC packet subelement entries loaded.")

        # TM packet structure in vpd SRDB data (Variable packet definition)
        self.add_tm_packetsubelement()
        logger.debug("TM packet subelement entries loaded.")

        # TM/TC static structure in plf SRDB data
        self.add_static_packetsubelement()
        logger.debug("Static packet subelement entries loaded.")

    def add_enum_palisade_ids(self):
        """
        Enumeration palisade id loading from txf, paf SRDB data (WIP)

        FIXME: load the palisade ids
        """

        # textual calibrations
        _ = self.from_table(
            os.path.join(self.dir_path, "txf.dat"), dtype=dtypes.txf_dtype
        )

        _ = self.from_table(
            os.path.join(self.dir_path, "paf.dat"), dtype=dtypes.paf_dtype
        )

    def add_tf_palisade_ids(self):
        """
        TF palisade id loading from caf, cca SRDB data

        FIXME: load the palisade ids
        """

        # numerical calibrations
        _ = self.from_table(
            os.path.join(self.dir_path, "caf.dat"), dtype=dtypes.caf_dtype
        )

        _ = self.from_table(
            os.path.join(self.dir_path, "cca.dat"), dtype=dtypes.cca_dtype
        )

    def create_transfer_functions(self):
        """
        TF points loading from cap, ccs SRDB data (tabulated transfer functions)
        """

        cap_table = self.from_table(
            os.path.join(self.dir_path, "cap.dat"),
            dtype=dtypes.cap_dtype,
            regex_dict={"NAME": "^CIW.*"},
        )

        ccs_table = self.from_table(
            os.path.join(self.dir_path, "ccs.dat"),
            dtype=dtypes.ccs_dtype,
            regex_dict={"NAME": "^CIW.*"},
        )

        self.tf_dict = {}

        for row in ccs_table:
            # parse the TC transfer functions
            srdbid = row["NAME"]
            self.tf_dict.setdefault(srdbid, []).append(
                {"raw": int(row["RAW"]), "eng": row["ENG"]}
            )
            logger.debug("Reading TF parameter value: %s" % row["NAME"])

        for row in cap_table:
            # parse the TM transfer functions
            srdbid = row["NAME"]
            self.tf_dict.setdefault(srdbid, []).append(
                {"raw": int(row["RAW"]), "eng": row["ENG"]}
            )
            logger.debug("Reading TF parameter value: %s" % row["NAME"])

        logger.debug("End of parsing TF")

    def create_enumerations(self):
        """
        Search in the txp and pas for enumeration definition.
        """

        txp_table = self.from_table(
            os.path.join(self.dir_path, "txp.dat"),
            dtype=dtypes.txp_dtype,
            regex_dict={"NAME": "^CIW.*"},
        )

        pas_table = self.from_table(
            os.path.join(self.dir_path, "pas.dat"),
            dtype=dtypes.pas_dtype,
            regex_dict={"NAME": "^CIW.*"},
        )

        enumerations = {}

        for row in txp_table:
            # get the TM info
            srdb_id = row["NAME"]
            enumerations.setdefault(srdb_id, {})[row["ENG"]] = int(row["RAW"])
            logger.debug("Reading TF parameter value: %s" % row["NAME"])

        for row in pas_table:
            # get the TM info
            srdb_id = row["NAME"]
            enumerations.setdefault(srdb_id, {})[row["ENG"]] = int(row["RAW"])
            logger.debug("Reading TF parameter value: %s" % row["NAME"])

        for srdb_id in enumerations:
            enumeration = Enumeration(srdb_id)
            enumeration.values = enumerations[srdb_id]

        logger.debug("End of parsing Enumerations")

    def create_global_parameters_mapping(self):
        pcf_glob_det_table = self.from_table(
            os.path.join(self.dir_path, "pcf_glob_det.dat"),
            dtype=dtypes.pcf_glob_det_dtype,
        )

        for row in pcf_glob_det_table:
            global_param_list = self.global_parameters.get(row["GLOBALID"])
            if global_param_list is None:
                global_param_list = []
            global_param_list.append(row["DETAILEDID"])
            self.global_parameters[row["GLOBALID"]] = global_param_list

    def parse(self):
        """
        Call all methods to get the necessary information from the SRDB .dat
        files.
        """
        logger.debug("Loading and parsing %s files ..." % self._source)

        # create mapping between detailed and global parameters
        self.create_global_parameters_mapping()

        # load min and max values of parameters
        self.load_min_max_values()

        # create enumerations
        self.create_enumerations()

        # create transfer functions
        self.create_transfer_functions()

        # create TM and TC parameters
        self.create_parameters()

        # create TM and TC packets
        self.create_packets()

        # populate TM and TC packets with parameters
        self.populate_packets_with_parameters()

        # remove numpy dtype for parameters and packets attributes
        for packet in Packet:
            drop_numpy_type(packet)
            for parameter in packet.parameters:
                drop_numpy_type(parameter)

        for definition in ParameterDefinition:
            drop_numpy_type(definition)
        logger.debug("Loading and parsing %s files ... Done" % self._source)



# -*- coding: utf-8 -*-
import numpy

# Note: we use object instead of int type where columns can contain empty values

__all__ = ["vdf_dtype", "sw_para_dtype", "cpc_dtype", "pcf_dtype", "cdf_dtype"]

vdf_dtype = numpy.dtype(
    [
        ("NAME", "U8"),  # short version name (8 char)
        ("COMMENT", "U32"),
        ("DOMAINID", int),
        ("RELEASE", int),
        ("ISSUE", int),
    ]
)

sw_para_dtype = numpy.dtype(
    [("SCOS_NAME", "U8"), ("SW_NAME", "U32"), ("SW_DESCR", "U255")]
)

cpc_dtype = numpy.dtype(
    [
        ("PNAME", "U8"),  # Complet SRDB ID (tm/tc + kind + srdbid)
        ("DESCR", "U24"),  # PALISADE ID
        ("PTC", int),
        ("PFC", int),
        ("DISPFMT", "U1"),
        ("RADIX", "U1"),
        ("UNIT", "U4"),
        ("CATEG", "U1"),
        ("PRFREF", "U10"),
        ("CCAREF", "U10"),
        ("PAFREF", "U10"),
        ("INTER", "U1"),
        ("DEFVAL", "U17"),
        # ('OBTID', int),
    ]
)

pcf_dtype = numpy.dtype(
    [
        ("NAME", "U8"),  # Complet SRDB ID (tm/tc + kind + srdbid)
        ("DESCR", "U24"),  # PALISADE ID
        ("PID", object),  # On-board  ID  of  the  telemetry  parameter
        ("UNIT", "U4"),
        ("PTC", int),  # Parameter Type Code.
        ("PFC", int),  # Parameter Format Code
        ("WIDTH", object),
        ("VALID", "U8"),
        ("RELATED", "U8"),
        ("CATEG", "U1"),
        ("NATUR", "U1"),
        ("CURTX", "U10"),  # Parameter calibration identification name
        ("INTER", "U1"),
        ("USCON", "U1"),
        ("DECIM", object),
        ("PARVAL", "U14"),
        ("SUBSYS", "U8"),
        ("VALPAR", object),
        ("SPTYPE", "U1"),
        # ('CORR', str), # Optional
        # ('OBTID', int),  # Optional
        # ('DARC', str),   # Optional
        # ('ENDIAN', str), # Optional
    ]
)

cdf_dtype = numpy.dtype(
    [
        ("CNAME", "U8"),  # Complete SRDB ID (tm/tc + kind + srdbid)
        ("ELTYPE", "U1"),
        ("DESCR", "U24"),  # PALISADE ID
        ("ELLEN", int),
        ("BIT", int),
        ("GRPSIZE", int),
        ("PNAME", "U8"),
        ("INTER", "U1"),
        ("VALUE", "U17"),
        ("TMID", "U8"),
    ]
)

plf_dtype = numpy.dtype(
    [
        ("NAME", "U8"),  # Complet SRDB ID (tm/tc + kind + srdbid)
        ("SPID", object),
        ("OFFBY", int),
        ("OFFBI", int),
        ("NBOCC", int),
        ("LGOCC", int),
        ("TIME", int),
        ("TDOCC", int),
    ]
)

cap_dtype = numpy.dtype(
    [
        ("NAME", "U10"),  # Complet SRDB ID
        ("RAW", object),  # Raw value
        ("ENG", "U30"),  # Eng Value
    ]
)

ccs_dtype = numpy.dtype(
    [
        ("NAME", "U10"),  # Complet SRDB ID
        ("ENG", "U30"),  # Eng Value
        ("RAW", int),  # Raw value
    ]
)

pas_dtype = numpy.dtype(
    [
        ("NAME", "U10"),  # Complet SRDB ID
        ("ENG", "U30"),  # Eng Value
        ("RAW", int),  # Raw value
    ]
)

txp_dtype = numpy.dtype(
    [
        ("NAME", "U10"),  # Complet SRDB ID
        ("RAW", "U14"),  # Raw value
        ("dum", "U14"),  # dummy value (identical to raw)
        ("ENG", "U14"),  # Eng Value
    ]
)

pcf_glob_det_dtype = numpy.dtype(
    [("GLOBALID", "U8"), ("DETAILEDID", "U8"), ("UNDEF", int)]
)

ccf_dtype = numpy.dtype(
    [
        ("CNAME", "U8"),  # Complet SRDB ID (tm/tc + kind + srdbid)
        ("DESCR", "U24"),  # PALISADE ID
        ("DESCR2", "U64"),  # Description
        ("CTYPE", "U8"),  # Command type
        ("CRITICAL", "U1"),  # Y/N
        ("PKTID", "U8"),  # Name  of  the  packet  header  (TCP_ID)
        ("TYPE", object),  # Service Type
        ("STYPE", object),  # Service Sub-Type
        (
            "APID",
            object,
        ),  # Integer value in the range (0....65535). PACKET_CATEGORY + PID.
        (
            "NPARS",
            object,
        ),  # Number of elements  (i.e. parameters or fixed areas)
        ("PLAN", "U1"),
        ("EXEC", "U1"),
        ("ILSCOPE", "U1"),
        ("ILSTAGE", "U1"),
        ("SUBSYS", object),
        ("HIPRI", "U1"),
        ("MAPID", object),
        ("DEFSET", "U8"),
        ("RAPID", object),
        ("ACK", object),
        ("SUBSCHED", object),
    ]
)

pid_dtype = numpy.dtype(
    [
        ("TYPE", object),
        ("STYPE", object),
        ("APID", int),
        ("PI1_VAL", object),  # SID
        ("PI2_VAL", object),
        ("SPID", object),
        ("DESCR", "U64"),
        ("UNIT", "U8"),
        ("TPSD", int),
        ("DFHSIZE", object),
        ("TIME", "U1"),
        ("INTER", object),
        ("VALID", "U1"),
        ("CHECK", object),
        ("EVENT", "U1"),
        ("EVID", "U17"),
    ]
)

vpd_dtype = numpy.dtype(
    [
        ("TPSD", int),
        ("POS", int),
        ("NAME", "U8"),
        ("GRPSIZE", int),
        ("FIXREP", int),
        ("CHOICE", "U1"),
        ("PIDREF", "U1"),
        ("DISDESC", "U16"),
        ("WIDTH", int),
        ("JUSTIFY", "U1"),
        ("NEWLINE", "U1"),
        ("DCHAR", int),
        ("FORM", "U1"),
        ("OFFSET", int),
    ]
)

caf_dtype = numpy.dtype(
    [
        ("NUMBR", "U10"),  # calibration identification name.
        ("DESCR", "U32"),  # palisade id
        ("ENGFMT", "U1"),
        ("RAWFMT", "U1"),
        ("RADIX", "U1"),
        ("UNIT", "U4"),
        ("NCURVE", int),
        (
            "INTER",
            "U1",
        ),  # Controls the extrapolation behaviour for raw values outside the range
    ]
)


# numerical (de-)calibration curves
cca_dtype = numpy.dtype(
    [
        ("NUMBR", "U10"),  # calibration identification name.
        ("DESCR", "U24"),  # palisade id
        ("ENGFMT", "U1"),
        ("RAWFMT", "U1"),
        ("RADIX", "U1"),
        ("UNIT", "U4"),
        ("NCURVE", int),
    ]
)


txf_dtype = numpy.dtype(
    [
        ("NUMBR", "U10"),  # calibration identification name.
        ("DESCR", "U32"),  # palisade id
        ("RAWFMT", "U1"),
        ("NALIAS", int),  # number of records
    ]
)


paf_dtype = numpy.dtype(
    [
        ("NUMBR", "U10"),  # calibration identification name.
        ("DESCR", "U24"),
        ("RAWFMT", "U1"),
        ("NALIAS", int),  # number of records
    ]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from poppy.core.logger import logger
from .base_palisade_parser import BasePALISADEParser

__all__ = [
    "PALISADEMetadataParser",
]


class PALISADEMetadataParser(BasePALISADEParser):
    """
    Class used for the parsing of the XML of the IDB (PALISADE) to get only the PALISADE metadata
    """

    def parse(self):
        """
        Call the load palisade method
        """

        self.load_palisade_metadata()

    def load_palisade_metadata(self):
        """
        Load the palisade metadata map

        :return:
        """

        # search for packets in the given node
        for packet_node in self._findall(
            self.root,
            ".//{0}PacketDefinition",
            "",
        ):
            # get the name of the packet
            palisade_id = packet_node.attrib["Name"]

            # search for the SRDB id of the packet
            if palisade_id[:2] not in ["TM", "TC"]:
                logger.warning("Skipping packet %s" % palisade_id)
                continue
            elif palisade_id in self.palisade_to_srdb_id[palisade_id[:2]]:
                # keep reference to mapping
                info = self.packets_mapping[palisade_id]

            else:
                logger.error(
                    (
                        "Packet {0} not found in PALISADE. Maybe an internal DPU"
                        + " command ?"
                    ).format(palisade_id)
                )
                continue

            # get the type of the packet
            packet_type_node = self._find(
                packet_node,
                ".//{0}Param[@StructParamIDRef='PACKET_TYPE']",
                "",
            )
            if packet_type_node is None:
                # the packet doesn't have a type, so don't analyze it
                logger.warning(
                    (
                        "Packet {0} doesn't have a type. Surely an internal "
                        + "command of the DPU."
                    ).format(palisade_id)
                )
                continue
            else:
                packet_type = packet_type_node.attrib["Value"][:2]

            # get the palisade category
            palisade_category = self.get_palisade_category(packet_node)

            # compute the long srdb id
            srdb_id = self.compute_packet_srdbid(packet_type, info["SRDBID"])

            # store the palisade id and the palisade category indexed by the srdb id
            self.palisade_metadata_map[srdb_id] = {
                "palisade_id": palisade_id,
                "palisade_category": palisade_category,
            }
